# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict

import numpy as np
import torch
from torch import nn

import nncf
from nncf.common.graph import NNCFNodeName
from nncf.torch.layer_utils import COMPRESSION_MODULES
from nncf.torch.layer_utils import StatefullModuleInterface


@COMPRESSION_MODULES.register()
class FilterPruningMask(nn.Module, StatefullModuleInterface):
    """
    A module contains the mask for pruning.
    On forward pass applying the mask to weight and bias of the module.
    """

    MASK_APPLYING_DIM_KEY = "dim"
    NODE_NAME_KEY = "node_name"
    SIZE_KEY = "size_key"

    def __init__(self, size, node_name, dim=0):
        super().__init__()
        self.register_buffer("_binary_filter_pruning_mask", torch.ones(size))
        self.mask_applying_dim = dim
        self.node_name = node_name

    @property
    def binary_filter_pruning_mask(self) -> torch.Tensor:
        return self._binary_filter_pruning_mask

    @binary_filter_pruning_mask.setter
    def binary_filter_pruning_mask(self, mask: torch.Tensor):
        with torch.no_grad():
            self._binary_filter_pruning_mask.set_(mask)

    def forward(self, **params):
        new_params = []
        for param_name, param_value in params.items():
            # In case of None weight (or bias) mask shouldn't be applied
            if param_value is None:
                new_params.append(param_value)
                continue

            # For weights self.mask_applying_dim should be used, for bias dim=0
            dim = 0 if param_name == "bias" else self.mask_applying_dim
            new_params.append(
                apply_filter_binary_mask(
                    self.binary_filter_pruning_mask, param_value, node_name_for_logging=self.node_name, dim=dim
                )
            )
        return new_params

    def get_config(self) -> Dict[str, Any]:
        return {
            self.MASK_APPLYING_DIM_KEY: self.mask_applying_dim,
            self.NODE_NAME_KEY: self.node_name,
            self.SIZE_KEY: list(self.binary_filter_pruning_mask.size()),
        }

    @classmethod
    def from_config(cls, state: Dict[str, Any]) -> "FilterPruningMask":
        return FilterPruningMask(
            size=state[cls.SIZE_KEY], node_name=state[cls.NODE_NAME_KEY], dim=state[cls.MASK_APPLYING_DIM_KEY]
        )


def broadcast_filter_mask(filter_mask, shape, dim=0):
    broadcasted_shape = np.ones(len(shape), dtype=np.int64)
    broadcasted_shape[dim] = filter_mask.size(0)
    broadcasted_filter_mask = torch.reshape(filter_mask, tuple(broadcasted_shape))
    return broadcasted_filter_mask


def apply_filter_binary_mask(
    filter_mask: torch.Tensor,
    module_parameter: torch.nn.Parameter,
    node_name_for_logging: NNCFNodeName = "",
    dim: int = 0,
):
    """
    Applying binary filter mask to parameter of the module - usually to weight/bias of convolution or linear layer.
    Mask is applied to a given dimension without overriding parameter's values.
    :param filter_mask: binary mask (should have the same shape as conv weight on the given dimension)
    :param module_parameter: a tensor representing a module parameter (e.g. weight or bias of convolution)
    :param node_name_for_logging: name of the module to which the mask is applied
    :param dim: a dimension to apply the mask (0 by default)
    :return: result with applied mask
    """
    if filter_mask.size(0) != module_parameter.size(dim):
        raise nncf.InternalError(
            "Shape of mask = {} for module {} isn't broadcastable to weight shape={}."
            " ".format(filter_mask.shape, node_name_for_logging, module_parameter.shape)
        )
    broadcasted_filter_mask = broadcast_filter_mask(filter_mask, module_parameter.shape, dim)
    return module_parameter.mul(broadcasted_filter_mask)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Tuple, Union

import nncf.torch.graph.operator_metatypes as om
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.graph.layer_attributes import ConstantLayerAttributes
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.layer_attributes import GenericWeightedLayerAttributes
from nncf.common.graph.layer_attributes import GetItemLayerAttributes
from nncf.common.graph.layer_attributes import GroupNormLayerAttributes
from nncf.common.graph.layer_attributes import LinearLayerAttributes
from nncf.common.graph.layer_attributes import MultipleInputLayerAttributes
from nncf.common.graph.layer_attributes import MultipleOutputLayerAttributes
from nncf.common.graph.layer_attributes import PadLayerAttributes
from nncf.common.graph.layer_attributes import PermuteLayerAttributes
from nncf.common.graph.layer_attributes import ReshapeLayerAttributes
from nncf.common.graph.layer_attributes import TransposeLayerAttributes
from nncf.common.graph.operator_metatypes import ConstNoopMetatype
from nncf.common.graph.operator_metatypes import get_all_aliases
from nncf.common.graph.utils import get_split_axis
from nncf.torch.dynamic_graph.trace_tensor import TracedParameter
from nncf.torch.layers import NNCF_MODULES_DICT

OP_NAMES_REQUIRING_MODULE_ATTRS = [v.op_func_name for v in NNCF_MODULES_DICT] + list(
    om.PTGroupNormMetatype.get_all_aliases()
)

TRANSPOSE_OP_NAMES = ["transpose", "transpose_"]
PERMUTE_OP_NAMES = ["permute"]
GETITEM_OP_NAMES = ["__getitem__"]
CONV_OP_NAMES = get_all_aliases(om.PTConv1dMetatype, om.PTConv2dMetatype, om.PTConv3dMetatype)
CONV_TRANSPOSE_OP_NAMES = get_all_aliases(
    om.PTConvTranspose1dMetatype, om.PTConvTranspose2dMetatype, om.PTConvTranspose3dMetatype
)
LINEAR_OP_NAMES = get_all_aliases(om.PTLinearMetatype)
BATCHNORM_OP_NAMES = get_all_aliases(om.PTBatchNormMetatype)
EMBEDDING_OP_NAMES = get_all_aliases(om.PTEmbeddingMetatype, om.PTEmbeddingBagMetatype)
GROUP_NORM_OP_NAMES = get_all_aliases(om.PTGroupNormMetatype)
LAYER_NORM_OP_NAMES = get_all_aliases(om.PTLayerNormMetatype)
PAD_OP_NAMES = om.PTPadMetatype.get_all_aliases()
CONCAT_OP_NAMES = om.PTCatMetatype.get_all_aliases()
CONST_OP_NAMES = ConstNoopMetatype.get_all_aliases()
OP_NAMES_REQUIRING_ATTRS_FROM_ARGS_KWARGS = list(
    TRANSPOSE_OP_NAMES + PERMUTE_OP_NAMES + GETITEM_OP_NAMES + PAD_OP_NAMES + CONCAT_OP_NAMES + CONST_OP_NAMES
)


def get_layer_attributes_from_args_and_kwargs(op_name: str, args, kwargs) -> BaseLayerAttributes:
    layer_attrs = None
    if op_name in CONV_OP_NAMES:
        layer_attrs = _get_conv_attrs_from_args_kwargs(args, kwargs)
    elif op_name in CONV_TRANSPOSE_OP_NAMES:
        layer_attrs = _get_conv_transpose_attrs_from_args_kwargs(args, kwargs)
    elif op_name in LINEAR_OP_NAMES:
        layer_attrs = _get_linear_attrs_from_args_kwargs(args, kwargs)
    elif op_name in GROUP_NORM_OP_NAMES:
        layer_attrs = _get_group_norm_attrs_from_args_kwargs(args, kwargs)
    elif op_name in BATCHNORM_OP_NAMES:
        layer_attrs = _get_batchnorm_attrs_from_args_kwargs(args, kwargs)
    elif op_name in LAYER_NORM_OP_NAMES:
        layer_attrs = _get_layer_norm_attrs_from_args_kwargs(args, kwargs)
    elif op_name in EMBEDDING_OP_NAMES:
        layer_attrs = _get_embedding_attrs_from_args_kwargs(args, kwargs)
    elif op_name in TRANSPOSE_OP_NAMES:
        layer_attrs = _get_transpose_attrs_from_args_kwargs(args, kwargs)
    elif op_name in PERMUTE_OP_NAMES:
        layer_attrs = _get_permute_attrs_from_args_kwargs(args, kwargs)
    elif op_name in GETITEM_OP_NAMES:
        layer_attrs = _get_getitem_attrs_from_args_kwargs(args, kwargs)
    elif op_name in PAD_OP_NAMES:
        layer_attrs = _get_pad_attrs_from_args_kwargs(args, kwargs)
    elif op_name in CONCAT_OP_NAMES:
        layer_attrs = _get_concat_attrs_from_args_kwargs(args, kwargs)
    elif op_name in CONST_OP_NAMES:
        layer_attrs = _get_const_attrs_from_args_kwargs(args, kwargs)
    return layer_attrs


def set_nodes_attributes_in_nncf_graph(graph: NNCFGraph) -> None:
    for node in graph.get_all_nodes():
        if node.metatype in [om.PTReshapeMetatype, om.PTSqueezeMetatype]:
            input_nodes = graph.get_input_edges(node)
            output_nodes = graph.get_output_edges(node)
            # In case ReshapeMetatype op is intermediate node
            if input_nodes and output_nodes:
                layer_attributes = ReshapeLayerAttributes(input_nodes[0].tensor_shape, output_nodes[0].tensor_shape)
                node.layer_attributes = layer_attributes

        if node.metatype is om.PTSplitMetatype:
            input_edges = graph.get_input_edges(node)
            output_edges = graph.get_output_edges(node)
            if input_edges and output_edges:
                input_shapes = [edge.tensor_shape for edge in input_edges]
                output_shapes = [edge.tensor_shape for edge in output_edges]
                axis = get_split_axis(input_shapes, output_shapes)
                chunks = len(output_edges)
                layer_attributes = MultipleOutputLayerAttributes(chunks, axis)
                node.layer_attributes = layer_attributes


def _get_transpose_attrs_from_args_kwargs(args, kwargs) -> TransposeLayerAttributes:
    return TransposeLayerAttributes(**_get_kwargs_shifted(["dim0", "dim1"], args, kwargs))


def _get_getitem_attrs_from_args_kwargs(args, kwargs):
    return GetItemLayerAttributes(key=args[1])


def _get_permute_attrs_from_args_kwargs(args, kwargs) -> PermuteLayerAttributes:
    arg_name = "dims"
    dims = kwargs[arg_name] if arg_name in kwargs else args[1:]
    return PermuteLayerAttributes(dims)


def _get_pad_attrs_from_args_kwargs(args, kwargs) -> PadLayerAttributes:
    mode = kwargs.get("mode", "constant" if len(args) < 3 else args[2])
    value = kwargs.get("value", 0 if len(args) < 4 else args[3])
    return PadLayerAttributes(mode, value)


def _get_concat_attrs_from_args_kwargs(args, kwargs) -> MultipleInputLayerAttributes:
    if "tensors" in kwargs:
        tensors = kwargs["tensors"]
    else:
        tensors = args[0]
    axis = kwargs.get("dim", 0 if len(args) < 2 else args[1])
    return MultipleInputLayerAttributes(axis=axis, num_inputs=len(tensors))


def _get_kwargs_shifted(args_names, args, kwargs, shift=1):
    res_kwargs = {}
    for idx, arg_name in enumerate(args_names):
        res_kwargs[arg_name] = kwargs[arg_name] if arg_name in kwargs else args[idx + shift]
    return res_kwargs


def _get_const_attrs_from_args_kwargs(args, _) -> ConstantLayerAttributes:
    name = "Unknown"
    shape = []
    if args and isinstance(args[0], TracedParameter):
        name = args[0].name
        shape = args[0].shape
    return ConstantLayerAttributes(name, shape)


def apply_args_defaults(
    args: List[Any], kwargs: Dict[str, Any], args_signature=List[Union[str, Tuple[str, Any]]]
) -> Dict[str, Any]:
    """
    Combines positional arguments (`args`) and keyword arguments (`kwargs`)
    according to the provided `args_signature`.

    The `args_signature` is a list that defines the expected arguments.
    Each element in the list can be either:

    - string: This represents the name of an argument expected to be a positional argument.
    - tuple: This represents the name and default value of an argument.
        - The first element in the tuple is the argument name.
        - The second element in the tuple is the default value.

    :param args: List of positional arguments.
    :param kwargs: Dictionary of keyword arguments.
    :param args_signature: List defining the expected arguments as described above.

    :return: A dictionary combining arguments from `args` and `kwargs` according to the `args_signature`.
    """
    # Manual defines function signature necessary because inspection of torch function is not available
    # https://github.com/pytorch/pytorch/issues/74539

    args_dict: Dict[str, Any] = dict()
    for idx, arg_desc in enumerate(args_signature):
        if isinstance(arg_desc, str):
            if arg_desc in kwargs:
                args_dict[arg_desc] = kwargs[arg_desc]
            elif idx < len(args):
                args_dict[arg_desc] = args[idx]
            else:
                raise ValueError("Incorrect args_signature, can not by applied to function arguments.")
        elif isinstance(arg_desc, Tuple):
            arg_name, default = arg_desc
            args_dict[arg_name] = kwargs.get(arg_name, args[idx] if idx < len(args) else default)
        else:
            raise ValueError("Incorrect args_signature, element of list should be str or tuple.")
    return args_dict


GENERIC_WEIGHT_FUNC_SIGNATURE = ["input", "weight"]
LINEAR_FUNC_SIGNATURE = ["input", "weight", ("bias", None)]
CONV_FUNC_SIGNATURE = ["input", "weight", ("bias", None), ("stride", 1), ("padding", 0), ("dilation", 1), ("groups", 1)]
CONV_TRANSPOSE_FUNC_SIGNATURE = [
    "input",
    "weight",
    ("bias", None),
    ("stride", 1),
    ("padding", 0),
    ("output_padding", 0),
    ("groups", 1),
    ("dilation", 1),
]
BATCH_NORM_FUNC_SIGNATURE = [
    "input",
    "running_mean",
    "running_var",
    ("weight", None),
    ("bias", None),
    ("training", False),
    ("momentum", 0.1),
    ("eps", 1e-5),
]
GROUP_NORM_FUNC_SIGNATURE = [
    "input",
    "num_groups",
    ("weight", None),
    ("bias", None),
    ("eps", None),
]
LAYER_NORM_FUNC_SIGNATURE = [
    "input",
    "normalized_shape",
    ("weight", None),
    ("bias", None),
    ("eps", 1e-05),
    ("cudnn_enable", True),
]


def _get_conv_attrs_from_args_kwargs(args: List[Any], kwargs: Dict[str, Any]) -> ConvolutionLayerAttributes:
    args_dict = apply_args_defaults(args, kwargs, CONV_FUNC_SIGNATURE)

    kernel_size = tuple(args_dict["weight"].shape[2:])
    in_channels = args_dict["weight"].shape[1] * args_dict["groups"]
    out_channels = args_dict["weight"].shape[0]

    return ConvolutionLayerAttributes(
        weight_requires_grad=args_dict["weight"].requires_grad,
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=kernel_size,
        stride=args_dict["stride"],
        dilations=args_dict["dilation"],
        groups=args_dict["groups"],
        transpose=False,
        padding_values=args_dict["padding"],
        with_bias=args_dict["bias"] is not None,
    )


def _get_conv_transpose_attrs_from_args_kwargs(args: List[Any], kwargs: Dict[str, Any]) -> ConvolutionLayerAttributes:
    args_dict = apply_args_defaults(args, kwargs, CONV_TRANSPOSE_FUNC_SIGNATURE)

    kernel_size = tuple(args_dict["weight"].shape[2:])
    in_channels = args_dict["weight"].shape[0]
    out_channels = args_dict["weight"].shape[1] * args_dict["groups"]

    return ConvolutionLayerAttributes(
        weight_requires_grad=args_dict["weight"].requires_grad,
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=kernel_size,
        stride=args_dict["stride"],
        dilations=args_dict["dilation"],
        groups=args_dict["groups"],
        transpose=True,
        padding_values=args_dict["padding"],
        with_bias=args_dict["bias"] is not None,
        output_padding_values=args_dict["output_padding"],
    )


def _get_linear_attrs_from_args_kwargs(args, kwargs) -> LinearLayerAttributes:
    args_dict = apply_args_defaults(args, kwargs, LINEAR_FUNC_SIGNATURE)
    return LinearLayerAttributes(
        weight_requires_grad=args_dict["weight"].requires_grad,
        in_features=args_dict["weight"].shape[1],
        out_features=args_dict["weight"].shape[0],
        with_bias=args_dict["bias"] is not None,
    )


def _get_batchnorm_attrs_from_args_kwargs(args, kwargs):
    args_dict = apply_args_defaults(args, kwargs, BATCH_NORM_FUNC_SIGNATURE)
    return GenericWeightedLayerAttributes(
        weight_requires_grad=False if args_dict["weight"] is None else args_dict["weight"].requires_grad,
        weight_shape=[] if args_dict["weight"] is None else args_dict["weight"].shape,
        filter_dimension_idx=0,
        with_bias=args_dict["bias"] is not None,
    )


def _get_embedding_attrs_from_args_kwargs(args, kwargs):
    args_dict = apply_args_defaults(args, kwargs, GENERIC_WEIGHT_FUNC_SIGNATURE)
    return GenericWeightedLayerAttributes(
        weight_requires_grad=args_dict["weight"].requires_grad,
        weight_shape=args_dict["weight"].shape,
        filter_dimension_idx=0,
        with_bias=False,
    )


def _get_group_norm_attrs_from_args_kwargs(args, kwargs):
    args_dict = apply_args_defaults(args, kwargs, GROUP_NORM_FUNC_SIGNATURE)
    return GroupNormLayerAttributes(
        weight_requires_grad=args_dict["weight"].requires_grad,
        num_channels=args_dict["weight"].shape[0],
        num_groups=args_dict["num_groups"],
    )


def _get_layer_norm_attrs_from_args_kwargs(args, kwargs):
    args_dict = apply_args_defaults(args, kwargs, LAYER_NORM_FUNC_SIGNATURE)
    return GenericWeightedLayerAttributes(
        weight_requires_grad=args_dict["weight"].requires_grad,
        weight_shape=args_dict["weight"].shape,
        filter_dimension_idx=0,
        with_bias=args_dict["bias"] is not None,
    )



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict, List, Optional, Set, Type

import nncf
from nncf.common.graph.definitions import NNCFGraphNodeType
from nncf.common.utils.registry import Registry


class OperatorMetatype:
    """
    Base class for grouping framework operators based on their semantic meaning.

    :param name: The name of the operator.
    :param hw_config_names: The names of the hardware configurations.
    :param output_channel_axis: The axis, along which the output channels of the operator are arranged.
    :param ignored_input_ports: Input ports of the operations that should not be considered for purposes of compression.
    """

    name: str = ""
    hw_config_names: List[str] = []
    output_channel_axis: Optional[int] = None
    ignored_input_ports: List[int] = []
    target_input_ports: Optional[List[int]] = None

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        """
        Returns a list of the framework operator aliases.

        :return: A list of the framework operator aliases.
        """
        return []

    @classmethod
    def get_subtypes(cls) -> List[Type["OperatorMetatype"]]:
        """
        Returns a list of 'OperatorMetatype' that are subtypes.

        :return: A subtype list.
        """
        return []

    @classmethod
    def subtype_check(cls, metatype: Type["OperatorMetatype"]) -> bool:
        """
        Check if a metatype is a subtype.

        :param metatype: An operator metatype.
        :return: True if metatype is a subtype otherwise False
        """
        subtypes = cls.get_subtypes()
        if metatype == cls or metatype in subtypes:
            return True

        return any(subtype.subtype_check(metatype) for subtype in subtypes)


class OperatorMetatypeRegistry(Registry):
    """
    Operator Metatypes Registry.
    """

    def __init__(self, name: str):
        """
        Initialize registry state.

        :param name: The registry name.
        """
        super().__init__(name)
        self._op_name_to_op_meta_dict: Dict[str, Type[OperatorMetatype]] = {}

    def register(self, name: Optional[str] = None, is_subtype: bool = False) -> Callable[..., Type[OperatorMetatype]]:
        """
        Decorator for registering operator metatypes.

        :param name: The registration name.
        :param is_subtype: Whether the decorated metatype is a subtype of another registered operator.
        :return: The inner function for registering operator metatypes.
        """
        name_ = name
        super_register = super()._register

        def wrap(obj: Type[OperatorMetatype]) -> Type[OperatorMetatype]:
            """
            Inner function for registering operator metatypes.

            :param obj: The operator metatype.
            :return: The input operator metatype.
            """
            cls_name = name_
            if cls_name is None:
                cls_name = obj.__name__
            super_register(obj, cls_name)
            if not is_subtype:
                op_names = obj.get_all_aliases()
                for name in op_names:
                    if name in self._op_name_to_op_meta_dict:
                        raise nncf.InternalError(
                            "Inconsistent operator metatype registry - single patched "
                            f"op name `{name}` maps to multiple metatypes!"
                        )
                    self._op_name_to_op_meta_dict[name] = obj
            return obj

        return wrap

    def get_operator_metatype_by_op_name(self, op_name: str) -> Type[OperatorMetatype]:
        """
        Returns the operator metatype by operator name.

        :param op_name: The operator name.
        :return: The operator metatype.
        """
        if op_name not in self._op_name_to_op_meta_dict:
            return UnknownMetatype
        return self._op_name_to_op_meta_dict[op_name]


NOOP_METATYPES = Registry("noop_metatypes")
INPUT_NOOP_METATYPES = Registry("input_noop_metatypes")
OUTPUT_NOOP_METATYPES = Registry("output_noop_metatypes")
CONST_NOOP_METATYPES = Registry("const_noop_metatypes")


class UnknownMetatype(OperatorMetatype):
    """
    UnknownMetatype is mapped to operations in NNCFGraph, which are unknown for algorithms,
    typically these are the operations that haven't been discovered before.
    Algorithms should avoid processing graph nodes with this metatype.
    """

    name = "unknown"

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        return [cls.name]


@NOOP_METATYPES.register()
class NoopMetatype(OperatorMetatype):
    """
    NoopMetatype is mapped to operations in NNCFGraph, that doesn't influence an input tensor.
    The compression algorithms can safely ignore this node.
    """

    name = "noop"

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        return [cls.name]


@NOOP_METATYPES.register()
@INPUT_NOOP_METATYPES.register()
class InputNoopMetatype(OperatorMetatype):
    name = "input_noop"

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        return [NNCFGraphNodeType.INPUT_NODE]


@NOOP_METATYPES.register()
@OUTPUT_NOOP_METATYPES.register()
class OutputNoopMetatype(OperatorMetatype):
    name = "output_noop"

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        return [NNCFGraphNodeType.OUTPUT_NODE]


@NOOP_METATYPES.register()
@CONST_NOOP_METATYPES.register()
class ConstNoopMetatype(OperatorMetatype):
    name = "const_noop"

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        return [NNCFGraphNodeType.CONST_NODE]


def get_all_aliases(*metatypes: OperatorMetatype) -> Set[str]:
    """
    Returns a set of all unique aliases from the provided metatypes.

    :param *metatypes: A list of operator metatypes.
    :return: A set containing all unique aliases for metatypes.
    """
    return set(a for m in metatypes for a in m.get_all_aliases())



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import shutil
from pathlib import Path

from nncf.common.quantization.quantizer_propagation.graph import QuantizerPropagationStateGraph
from nncf.common.quantization.quantizer_propagation.solver import QuantizerPropagationSolver


class QuantizerPropagationVisualizer:
    """
    An object performing visualization of the quantizer propagation algorithm's state into a chosen directory.
    """

    def __init__(self, dump_dir: str = None):
        self.dump_dir = Path(dump_dir)
        if self.dump_dir.exists():
            shutil.rmtree(str(self.dump_dir))

    def visualize_quantizer_propagation(
        self, prop_solver: QuantizerPropagationSolver, prop_graph: QuantizerPropagationStateGraph, iteration: str
    ):
        self.dump_dir.mkdir(parents=True, exist_ok=True)
        fname = "quant_prop_iter_{}.dot".format(iteration)
        prop_solver.debug_visualize(prop_graph, str(self.dump_dir / Path(fname)))



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, List

from nncf.common.logging import nncf_logger
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_dim import ElasticityDim

DEFAULT_STAGE_LR_RATE = 3.5e-06


class SDescriptorParamNames:
    TRAIN_DIMS = "train_dims"
    EPOCHS = "epochs"
    REORG_WEIGHTS = "reorg_weights"
    WIDTH_INDICATOR = "width_indicator"
    DEPTH_INDICATOR = "depth_indicator"
    BN_ADAPT = "bn_adapt"
    INIT_LR = "init_lr"
    EPOCHS_LR = "epochs_lr"
    SAMPLE_RATE = "sample_rate"


class StageDescriptor:
    """
    Describes parameters of the training stage. The stage defines active elastic dimension and its parameters.
    """

    _state_names = SDescriptorParamNames

    def __init__(
        self,
        train_dims: List[ElasticityDim],
        epochs: int = 1,
        reorg_weights: bool = False,
        bn_adapt: bool = False,
        depth_indicator: int = 1,
        width_indicator: int = 1,
        init_lr: float = None,
        epochs_lr: int = None,
        sample_rate: int = 1,
    ):
        self.train_dims = train_dims
        self.epochs = epochs
        self.depth_indicator = depth_indicator
        self.width_indicator = width_indicator
        self.reorg_weights = reorg_weights
        self.bn_adapt = bn_adapt
        self.init_lr = init_lr
        self.epochs_lr = epochs_lr
        self.sample_rate = sample_rate
        if sample_rate <= 0:
            nncf_logger.warning(f"Only positive integers are allowed for sample rate, but sample_rate={sample_rate}.")
            nncf_logger.warning("Setting sample rate to default 1")
            self.sample_rate = 1

    def __eq__(self, other: "StageDescriptor"):
        return self.__dict__ == other.__dict__

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "StageDescriptor":
        """
        Creates the object from its config.
        """
        train_dims = config.get(cls._state_names.TRAIN_DIMS, ["kernel"])
        kwargs = {
            cls._state_names.TRAIN_DIMS: [ElasticityDim(dim) for dim in train_dims],
            cls._state_names.EPOCHS: config.get(cls._state_names.EPOCHS, 1),
            cls._state_names.REORG_WEIGHTS: config.get(cls._state_names.REORG_WEIGHTS, False),
            cls._state_names.WIDTH_INDICATOR: config.get(cls._state_names.WIDTH_INDICATOR, 1),
            cls._state_names.DEPTH_INDICATOR: config.get(cls._state_names.DEPTH_INDICATOR, 1),
            cls._state_names.BN_ADAPT: config.get(cls._state_names.BN_ADAPT, False),
            cls._state_names.INIT_LR: config.get(cls._state_names.INIT_LR, None),
            cls._state_names.EPOCHS_LR: config.get(cls._state_names.EPOCHS_LR, None),
            cls._state_names.SAMPLE_RATE: config.get(cls._state_names.SAMPLE_RATE, 1),
        }
        return cls(**kwargs)

    @classmethod
    def from_state(cls, state: Dict[str, Any]):
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        kwargs = state.copy()
        train_dims = state[cls._state_names.TRAIN_DIMS]
        kwargs[cls._state_names.TRAIN_DIMS] = [ElasticityDim(dim) for dim in train_dims]
        return cls(**kwargs)

    def get_state(self) -> Dict[str, Any]:
        state_dict = {
            self._state_names.TRAIN_DIMS: [dim.value for dim in self.train_dims],
            self._state_names.EPOCHS: self.epochs,
            self._state_names.REORG_WEIGHTS: self.reorg_weights,
            self._state_names.WIDTH_INDICATOR: self.width_indicator,
            self._state_names.DEPTH_INDICATOR: self.depth_indicator,
            self._state_names.BN_ADAPT: self.bn_adapt,
            self._state_names.SAMPLE_RATE: self.sample_rate,
        }
        if self.init_lr is not None:
            state_dict["init_lr"] = self.init_lr
        if self.epochs_lr is not None:
            state_dict["epochs_lr"] = self.epochs_lr
        return state_dict



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, TypeVar

import numpy as np

import nncf
from nncf import Dataset
from nncf import nncf_logger
from nncf.common.factory import EngineFactory
from nncf.common.factory import ModelTransformerFactory
from nncf.common.factory import NNCFGraphFactory
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFGraphEdge
from nncf.common.graph import NNCFNode
from nncf.common.graph.definitions import NNCFGraphNodeType
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.logging.track_progress import track
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import copy_model
from nncf.common.utils.backend import get_backend
from nncf.quantization.algorithms.algorithm import Algorithm

TModel = TypeVar("TModel")

BIAS_CORRECTION_THRESHOLD = 1000
OUTPUT_PORT_OF_NODE = 0


class BiasCorrection(Algorithm):
    """
    Post-training BiasCorrection algorithm implementation

    The main purpose of this algorithm to reduce quantization error
    via correction the bias of the Convolutions, FullyConnected, etc. layers.
    The algorithm's pipeline looks like this:
        - we collect floating-point statistics from the first layers with the bias in the model;
        - then we get the quantized model and collect the necessary information to create a sub-graph;
        - after the information collection, we drop the quantizer-dequantizer pair or fake quantize node on activations;
        - using collected information, we try to create the model sub-graph dynamically:
        this sub-graph contains the layer in which bias would be corrected and the other layers that need to
        collect the new statistics for the next layer with the bias;
        - the shift calculates using the sub-graph that consists of the correction layer and
        weight quantizer-dequantizer pair or fake quantize node, and some other layers;
        - then we correct the original bias by the difference (shift) between floating-point and quantized outputs in
        the sub-graph and the model without quantizer-dequantizer pair or fake quantize node.
        - at the next step, we collect the new statistics for the next layer (that would be corrected) from
        the sub-graph with the updated bias value on the current step;
        - after the new statistics were collected, we drops the unnecessary statistics to reduce memory consumption;
        - in the end, we correct all needed biases in the original model.
    """

    def __init__(
        self,
        subset_size: int = 100,
        threshold: float = BIAS_CORRECTION_THRESHOLD,
        apply_for_all_nodes: bool = False,
        inplace_statistics: bool = True,
        backend_params: Optional[Dict[str, Any]] = None,
    ):
        """
        :param subset_size: Size of a subset for the statistics collection,
            defaults to 100.
        :param threshold: The magnitude threshold that regulates the application of the
            shift. Magnitude calculates as the maximum of the absolute ratio of the
            shift to the original bias value. If the calculated value is less than the
            threshold, the shift will apply to the bias, defaults to 1000.
        :param apply_for_all_nodes: If True, then the bias correction be applied to all
            quantized nodes, if the node has no bias then a bias node will be inserted,
            and if False, then the bias correction will only be applied to quantized
            nodes that have a bias.
        :param inplace_statistics: Defines wheather to calculate quantizers statistics
            by backend graph operations or by default Python implementation, defaults
            to True.
        :param backend_params: Backend specific parameters.
        """
        super().__init__()
        self.subset_size = subset_size
        self.threshold = threshold
        self.apply_for_all_nodes = apply_for_all_nodes
        self.inplace_statistics = inplace_statistics
        self.backend_params = backend_params
        self.nncf_graph = None
        self._backend_entity = None
        self._collected_stat_inputs_map = {}
        self._fp_inputs = defaultdict(list)
        self._algorithm_key = f"BC_{hash(self)}"

        if self.apply_for_all_nodes:
            raise nncf.InternalError("BiasCorrection algorithm does not support apply_for_all_nodes=True yet")

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.ONNX, BackendType.OPENVINO]

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic of the algorithm.

        :param model: Backend-specific input model.
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.ONNX:
            from nncf.quantization.algorithms.bias_correction.onnx_backend import ONNXBiasCorrectionAlgoBackend

            self._backend_entity = ONNXBiasCorrectionAlgoBackend()
        elif model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.bias_correction.openvino_backend import OVBiasCorrectionAlgoBackend

            self._backend_entity = OVBiasCorrectionAlgoBackend()
        else:
            raise nncf.UnsupportedBackendError(
                "Cannot return backend-specific entity because {} is not supported!".format(model_backend.value)
            )

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        self._set_backend_entity(model)
        main_transformations_layout = TransformationLayout()
        main_model_transformer = ModelTransformerFactory.create(model)

        model_copy = copy_model(model)
        graph_copy = NNCFGraphFactory.create(model_copy)
        model_copy = self._backend_entity.remove_fq_from_inputs(model_copy, graph_copy)
        nncf_graph = NNCFGraphFactory.create(model_copy)

        nodes_with_bias = []
        for node in nncf_graph.topological_sort():
            if self._is_node_correctable(node, nncf_graph):
                nodes_with_bias.append(node)

        # We pre-collect information about the subgraph we need in order
        # to collect statistics for the change in the bias of each layer.
        # Also here we collect a list of layers that depend on the current one.

        # The collected information contains lists of input and output layers,
        # for which we will create a subgraph for inference and collection of statistics.
        subgraphs_data = [self._get_subgraph_data_for_node(node, nncf_graph) for node in nodes_with_bias]

        for position, (node, subgraph_data) in track(
            list(enumerate(zip(nodes_with_bias, subgraphs_data))), description="Applying Bias correction"
        ):
            node_name = node.node_name

            # We do not make an additional copy of the model because
            # the model transformer (that uses during sub-graph extraction) already does this internally when creating.
            model_copy_subgraph = self._prepare_subgraph(node, model_copy, nncf_graph, subgraph_data)

            # Then we create the necessary data lists from the previously collected statistics,
            # for the subgraph inference.
            feed_dicts = self._create_feed_dicts(model_copy_subgraph, subgraph_data, statistic_points)

            bias_shift = self._compute_bias_shift(node, model_copy_subgraph, feed_dicts, statistic_points)

            current_bias = self._backend_entity.get_bias_value(node, model_copy, nncf_graph)

            channel_axis = node.metatype.output_channel_axis
            if current_bias.ndim > 1:
                channel_axis = range(current_bias.ndim)[channel_axis]
                axes = [i for i in range(current_bias.ndim) if i != channel_axis]
                bias_shift = np.expand_dims(bias_shift, axes)

            updated_bias = current_bias + bias_shift
            magnitude = self._get_bias_shift_magnitude(current_bias, updated_bias)

            if magnitude < self.threshold:
                nncf_logger.debug(f"{node_name} bias would be changed. Magnitude: {magnitude}")
                bias_correction_command = self._backend_entity.create_bias_correction_command(
                    node, updated_bias, nncf_graph
                )
                model_copy_subgraph = self._correct_bias(model_copy_subgraph, bias_correction_command)
                model_copy = self._correct_bias(model_copy, bias_correction_command)
                main_transformations_layout.register(bias_correction_command)
            else:
                nncf_logger.debug(f"{node_name} bias skipped by threshold. Magnitude: {magnitude}")

            # After collecting data to change the bias value, we need to collect statistics for subsequent nodes,
            # but already take into account the bias update made earlier.
            self._collect_new_stats(model_copy_subgraph, feed_dicts, subgraph_data)

            # Also, we need to remove unnecessary statistics that we don't need anymore,
            # to reduce memory usage during the algorithm's pipeline.
            self._remove_unnecessary_stats(position, subgraphs_data)

        return main_model_transformer.transform(main_transformations_layout)

    def _is_node_correctable(self, node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        """
        Verify if node bias can be corrected or not.

        :param node: NNCFNode instance to verify.
        :param nncf_graph: NNCFGraph instance.
        :return: Boolean value as the result.
        """
        return self._backend_entity.is_node_with_bias(node, nncf_graph) and self._backend_entity.is_quantized_weights(
            node, nncf_graph
        )

    def _get_subgraph_data_for_node(self, node: NNCFNode, nncf_graph: NNCFGraph) -> Dict[str, List[str]]:
        """
        This method collects necessary data for the specified node and its subgraph.
        This data contains the nodes (NNCFNode) for the subgraph building
        and statistics collection (for the next step).

        :param node: NNCFNode instance. This is the main node with bias that would be corrected (or not).
        :param nncf_graph: NNCFGraph instance for graph analysis.
        :return: A dict with the list of the nodes for the subgraph input and statistics collection.
        """
        statistic_nodes = []
        subgraph_input_ids, subgraph_output_ids = [], []

        def fill_subgraph_output_ids(edge: NNCFGraphEdge):
            node = edge.to_node
            if node in visited_nodes:
                return
            visited_nodes.append(node)

            input_id = (node.node_name, edge.input_port_id)
            output_id = (edge.from_node.node_name, edge.output_port_id)

            # If we found a node with bias, we have to collect it as a statistic node,
            # and its input for _collected_stat_inputs_map,
            # which will be used during the collection of statistics for the next node.
            if self._is_node_correctable(node, nncf_graph):
                subgraph_output_ids.append(output_id)
                self._collected_stat_inputs_map[input_id] = output_id

                statistic_nodes.append(edge.from_node)
                return

            for output_edge in nncf_graph.get_output_edges(node):
                fill_subgraph_output_ids(output_edge)

        def fill_subgraph_input_ids(edge: NNCFGraphEdge):
            node = edge.from_node
            input_id = (edge.to_node.node_name, edge.input_port_id)

            # Since we need to find the inputs for the subgraph,
            # we can take only those layers for which we have already collected statistics.
            if input_id in self._collected_stat_inputs_map:
                activation_id = self._collected_stat_inputs_map[input_id]
                if activation_id not in subgraph_output_ids:
                    subgraph_input_ids.append(input_id)
                    return

            if node in visited_nodes:
                return
            visited_nodes.append(node)

            for input_edge in nncf_graph.get_input_edges(node):
                fill_subgraph_input_ids(input_edge)

        # First, we need to find out the nodes with bias that follow by main node.
        # To collect statistics for next nodes.
        visited_nodes = []
        for output_edge in nncf_graph.get_output_edges(node):
            fill_subgraph_output_ids(output_edge)

        # We then need to find nodes for which statistics have already been collected,
        # to use them as inputs for the subgraph.
        statistic_nodes = statistic_nodes if statistic_nodes else nncf_graph.get_next_nodes(node)
        visited_nodes = []
        for stat_node in statistic_nodes:
            for input_edge in nncf_graph.get_input_edges(stat_node):
                fill_subgraph_input_ids(input_edge)

        if not subgraph_output_ids:
            for edge in nncf_graph.get_output_edges(node):
                next_node_name = edge.to_node.node_name
                if NNCFGraphNodeType.OUTPUT_NODE not in next_node_name:
                    subgraph_output_ids.append((edge.to_node.node_name, OUTPUT_PORT_OF_NODE))

        # In case the outputs were not found during the collection of statistics nodes,
        # we use the latter as the outputs of the subgraph.
        subgraph_data = {
            "subgraph_input_ids": set(subgraph_input_ids),
            "subgraph_output_ids": set(subgraph_output_ids),
        }

        return subgraph_data

    def _prepare_subgraph(self, node: NNCFNode, model: TModel, nncf_graph: NNCFGraph, subgraph_data: Dict) -> TModel:
        """
        This method prepares the subgraph from the model for the further inference.

        :param node: NNCFNode instance for the current layer.
        :param model: Backend-specific model instance.
        :param nncf_graph: Instance of NNCFGraph.
        :param subgraph_data: A dictionary with the layers for the graph building.
        :return: Backend-specific subgraph extracted from the model.
        """
        extracted_model = self.extract_model(
            model, subgraph_data["subgraph_input_ids"], subgraph_data["subgraph_output_ids"]
        )

        transformation_layout = TransformationLayout()
        model_transformer = ModelTransformerFactory.create(extracted_model)

        # For layers with weights, there is only one output port - 0.
        statistic_point = self._backend_entity.target_point(
            TargetType.POST_LAYER_OPERATION, node.node_name, port_id=OUTPUT_PORT_OF_NODE
        )
        output_insertion_command = self._backend_entity.output_insertion_command(nncf_graph, statistic_point)
        transformation_layout.register(output_insertion_command)
        return model_transformer.transform(transformation_layout)

    def _create_feed_dicts(
        self, model: TModel, subgraph_data: Dict, statistic_points: StatisticPointsContainer
    ) -> List[Dict]:
        """
        Creates the list of the dictionaries that contains the input data for the model execution.

        :param model: TModel instance.
        :param subgraph_data: A dictionary with the necessary data for current node.
        :param statistic_points: StatisticPointsContainer instance.
        :return: List of the dictionaries with the input data.
        """
        feed_dicts = []
        statistics_size = self.subset_size
        statistics_per_input = {}

        for input_node_name, input_port_id in subgraph_data["subgraph_input_ids"]:
            input_tensor_name = self._backend_entity.get_input_name(model, input_node_name, input_port_id)
            activation_name, output_port_id = self._collected_stat_inputs_map[(input_node_name, input_port_id)]
            input_fp = self._get_fp_inputs(statistic_points, node_name=activation_name, port_id=output_port_id)
            statistics_per_input[input_tensor_name] = input_fp
            statistics_size = min(statistics_size, len(input_fp))

        for stat_id in range(statistics_size):
            feed_dict = {}
            for input_node_name, input_port_id in subgraph_data["subgraph_input_ids"]:
                input_tensor_name = self._backend_entity.get_input_name(model, input_node_name, input_port_id)
                # Since we do not use as inputs the layers from which the statistics are gathered,
                # but those that follow them, we need to take this into account when creating feed dicts.
                activation_name, _ = self._collected_stat_inputs_map[(input_node_name, input_port_id)]
                feed_dict[input_tensor_name] = statistics_per_input[input_tensor_name][stat_id]
            feed_dicts.append(feed_dict)
        return feed_dicts

    def _compute_bias_shift(
        self, node: NNCFNode, model: TModel, feed_dicts: List, statistic_points: StatisticPointsContainer
    ) -> np.ndarray:
        """
        Computes bias shift that will be used for the further bias correction.

        :param node: NNCFNode instance, current layer.
        :param model: Backend-specific model.
        :param feed_dicts: List of dictionaries with the input data for model execution.
        :param statistic_points: StatisticPointsContainer instance.
        :return: Calculated bias shift value.
        """
        output_fp = self._get_fp_outputs(statistic_points, node.node_name)
        output_tensor_name = self._backend_entity.get_output_name(model, node.node_name, OUTPUT_PORT_OF_NODE)
        engine = EngineFactory.create(model)
        channel_axis = node.metatype.output_channel_axis
        q_outputs = []
        for feed_dict in feed_dicts:
            q_output = engine.infer(feed_dict)
            q_output = self._backend_entity.process_model_output(q_output, output_tensor_name)
            q_outputs.append(self._backend_entity.tensor_processor.mean_per_channel(q_output, channel_axis).tensor)
        # Here we get the per-sample average, so the axis is 0.
        q_output = np.mean(q_outputs, axis=0)
        return output_fp - q_output

    @staticmethod
    def _get_bias_shift_magnitude(current_bias_value: np.ndarray, updated_bias_value: np.ndarray) -> float:
        """
        Calculates bias shift magnitude based on the current and updated values.

        :param current_bias_value: Initial bias value.
        :param updated_bias_value: Updated bias value.
        :return: Magnitude between original and updated bias values.
        """
        bias_shift_magnitude = np.inf
        bias_shift_magnitude = np.max(
            np.abs(
                (updated_bias_value - current_bias_value)
                / (current_bias_value + np.finfo(current_bias_value.dtype).min)
            )
        )
        return bias_shift_magnitude

    def _correct_bias(self, model: TModel, bias_correction_command: TransformationCommand) -> TModel:
        """
        Returns the model (which can be represented as subgraph) with the updated bias value for the current layer.

        :param model: Backend-specific model.
        :param bias_correction_command: TransformationCommand instance for the bias correction.
        :return: Backend-specific model, but with the updated bias value.
        """
        model_transformer = ModelTransformerFactory.create(model)
        transformation_layout = TransformationLayout()
        transformation_layout.register(bias_correction_command)
        return model_transformer.transform(transformation_layout)

    def _collect_new_stats(self, model: TModel, feed_dicts: List, subgraph_data: Dict) -> None:
        """
        Updates the self._fp_inputs with the new statistics for the next layers
        after the correction of the bias for the current.

        :param model: Backend-specific subgraph.
        :param feed_dicts: List of dictionaries with the input data for the subgraph.
        :param subgraph_data: A dictionary with the needed list of the statistic nodes that will be updated.
        """
        engine = EngineFactory.create(model)
        for feed_dict in feed_dicts:
            new_q_output = engine.infer(feed_dict)
            for output_node_name, output_id in subgraph_data["subgraph_output_ids"]:
                output_tensor_name = self._backend_entity.get_output_name(model, output_node_name, output_id)
                self._fp_inputs[(output_node_name, output_id)].append(new_q_output[output_tensor_name])

    def _remove_unnecessary_stats(self, position: int, subgraphs_data: Dict[str, Dict]) -> None:
        """
        Removes unnecessary statistics that were collected before to reduce the memory usage.

        :param position: Zero-based position of the current node that was corrected.
        :param subgraphs_data: A dictionary of the data (input & statistic node names) that
            uses for the sub-graphs creation.
        """
        # Collects list of the statistics that needed for the future layers.
        needed_stats_list = []
        for i in range(position + 1, len(subgraphs_data)):
            input_ids = subgraphs_data[i]["subgraph_input_ids"]
            needed_stats_list.extend([self._collected_stat_inputs_map[input_id][0] for input_id in input_ids])

        node_inputs_ids = subgraphs_data[position]["subgraph_input_ids"]
        for node_input_id in node_inputs_ids:
            activation_name, port_id = self._collected_stat_inputs_map[node_input_id]
            input_id = (activation_name, port_id)
            if activation_name not in needed_stats_list and input_id in self._fp_inputs:
                nncf_logger.debug(f"Dropped {activation_name} output statistics.")
                self._fp_inputs[input_id] = []

    def _get_fp_inputs(self, statistic_points: StatisticPointsContainer, node_name: str, port_id: int) -> np.ndarray:
        """
        # TODO: paraphrase
        Makes out pre-layer needed data from the floating-point collected statistics.

        :param statistic_points: Filled StatisticPointsContainer.
        :param node_name: Name of the current layer.
        :param port_id: Port id for statistics collection.
        :return: Collected mean tensor data and shape for the further bias calculation.
        """

        def input_filter_func(point):
            # For the floating-point statistics collected in POST_LAYER style,
            # we also need to determine the output port id.
            # For the cases when the layer has more than one (0) output port.
            return (
                self._algorithm_key in point.algorithm_to_tensor_collectors
                and point.target_point.type == TargetType.POST_LAYER_OPERATION
                and point.target_point.port_id == port_id
            )

        input_id = (node_name, port_id)
        if input_id in self._fp_inputs:
            return self._fp_inputs[input_id]

        input_fp = []
        for tensor_collector in statistic_points.get_algo_statistics_for_node(
            node_name, input_filter_func, self._algorithm_key
        ):
            input_fp.extend(tensor_collector.get_statistics().values)
        self._fp_inputs[input_id] = input_fp
        return self._fp_inputs[input_id]

    def _get_fp_outputs(self, statistic_points: StatisticPointsContainer, node_name: str) -> np.ndarray:
        """
        Makes out post-layer needed data from the floating-point collected statistics.

        :param statistic_points: Filled StatisticPointsContainer.
        :param node_name: Name of the current layer.
        :return: Collected mean tensor data for the further bias calculation.
        """

        def output_filter_func(point):
            return (
                self._algorithm_key in point.algorithm_to_tensor_collectors
                and point.target_point.type == TargetType.POST_LAYER_OPERATION
            )

        output_fp = []
        for tensor_collector in statistic_points.get_algo_statistics_for_node(
            node_name, output_filter_func, self._algorithm_key
        ):
            output_fp.extend(tensor_collector.get_statistics().mean_values)
        return np.array(output_fp)

    def get_statistic_points(self, model: TModel, graph: NNCFGraph) -> StatisticPointsContainer:
        self._set_backend_entity(model)
        statistic_container = StatisticPointsContainer()

        nodes_with_bias = [
            node for node in graph.topological_sort() if self._backend_entity.is_node_with_bias(node, graph)
        ]
        model_inputs = graph.get_input_nodes()

        # Collection of statistics after layers where biases will be corrected.
        for node in nodes_with_bias:
            node_name = node.node_name
            channel_axis = node.metatype.output_channel_axis

            # For layers with weights, there is only one output port - 0.
            statistic_point = self._backend_entity.target_point(
                TargetType.POST_LAYER_OPERATION, node_name, port_id=OUTPUT_PORT_OF_NODE
            )
            stat_collector = self._backend_entity.mean_statistic_collector(
                channel_axis=channel_axis, num_samples=self.subset_size, inplace=self.inplace_statistics
            )
            statistic_container.add_statistic_point(
                StatisticPoint(
                    target_point=statistic_point, tensor_collector=stat_collector, algorithm=self._algorithm_key
                )
            )

        # We must collect the nodes with biases following the model inputs.
        biased_after_input_nodes = self._get_biased_after_nodes(graph, model_inputs, model)

        for biased_after_input_node in biased_after_input_nodes:
            # We need to collect activation input to register it for the biased layer as the layer with statistics.
            activation_port = self._backend_entity.get_activation_port_id(node, graph)
            edge = graph.get_input_edges(biased_after_input_node)[activation_port]

            input_id = (biased_after_input_node.node_name, edge.input_port_id)
            output_id = (edge.from_node.node_name, edge.output_port_id)

            if edge.from_node.node_name in statistic_container:
                continue

            self._collected_stat_inputs_map[input_id] = output_id
            statistic_point = self._backend_entity.target_point(
                TargetType.POST_LAYER_OPERATION, edge.from_node.node_name, port_id=edge.output_port_id
            )
            stat_collector = self._backend_entity.raw_statistic_collector(num_samples=self.subset_size)
            statistic_container.add_statistic_point(
                StatisticPoint(
                    target_point=statistic_point, tensor_collector=stat_collector, algorithm=self._algorithm_key
                )
            )

        # Then we need also to collect model input statistics to prevent cases when nodes with bias have no input data.
        for input_node in model_inputs:
            # We need to map all input node outputs as nodes with statistic
            for edge in graph.get_output_edges(input_node):
                input_id = (edge.to_node.node_name, edge.input_port_id)
                output_id = (input_node.node_name, edge.output_port_id)
                self._collected_stat_inputs_map[input_id] = output_id

            if input_node.node_name in statistic_container:
                continue

            statistic_point = self._backend_entity.target_point(
                TargetType.POST_LAYER_OPERATION, input_node.node_name, port_id=OUTPUT_PORT_OF_NODE
            )
            stat_collector = self._backend_entity.raw_statistic_collector(num_samples=self.subset_size)
            statistic_container.add_statistic_point(
                StatisticPoint(
                    target_point=statistic_point, tensor_collector=stat_collector, algorithm=self._algorithm_key
                )
            )

        return statistic_container

    def _get_biased_after_nodes(self, nncf_graph: NNCFGraph, nodes: List[NNCFNode], model: TModel) -> List[NNCFNode]:
        """
        This method finds and returns nodes with the bias in the model that follows after the input nodes.

        :param nncf_graph: NNCFGraph instance.
        :param nodes: List of the model inputs as NNCFNodes.
        :param model: TModel instance.
        :return: List of the nodes with bias.
        """

        def traverse_to_biased(node, condition_container):
            # A small hack to speed up graph traversal.
            if node in visited_nodes:
                return
            visited_nodes.append(node)

            # We need to collect nodes for the next recursion step.
            node_children = nncf_graph.get_next_nodes(node)

            # Check that node is with bias.
            if self._backend_entity.is_node_with_bias(node, nncf_graph):
                condition_container.add(node)
                return

            for node_child in node_children:
                traverse_to_biased(node_child, condition_container)

        biased_nodes = set()
        visited_nodes = []
        for node in nodes:
            nncf_logger.debug(f"Looking for biased nodes after {node.node_name} layer.")
            traverse_to_biased(node, condition_container=biased_nodes)

        dependant_nodes = set()
        # After finding the nodes following the provided layers, we need to make sure
        # that the found nodes really only depend on the main layers, and not on each other.
        for biased_node in biased_nodes:
            visited_nodes = []
            nncf_logger.debug(f"Filtering biased nodes after {biased_node.node_name} layer.")
            for next_node in nncf_graph.get_next_nodes(biased_node):
                traverse_to_biased(next_node, condition_container=dependant_nodes)

        return list(biased_nodes - dependant_nodes)

    def extract_model(
        self, model: TModel, input_node_ids: List[Tuple[str, int]], output_node_ids: List[Tuple[str, int]]
    ) -> TModel:
        """
        Returns the backend-specific model that bounded by the specified input & output layers.

        :param model: Backend-specific model.
        :param input_node_ids: List with the input node IDs.
        :param output_node_ids: List with the output node IDs.
        :return: Extracted backend-specific model.
        """
        transformation_layout = TransformationLayout()
        model_transformer = ModelTransformerFactory.create(model)
        model_extraction_command = self._backend_entity.model_extraction_command(input_node_ids, output_node_ids)
        transformation_layout.register(model_extraction_command)
        return model_transformer.transform(transformation_layout)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import List, Optional, Tuple, TypeVar

import numpy as np

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.tensor import NNCFTensor
from nncf.common.tensor_statistics.collectors import TensorStatisticCollectorBase

TModel = TypeVar("TModel")
OutputType = TypeVar("OutputType")


class BiasCorrectionAlgoBackend(ABC):
    @property
    @abstractmethod
    def tensor_processor(self):
        """
        Returns backend-specific instance of the NNCFCollectorTensorProcessor.
        """

    @staticmethod
    @abstractmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> TargetPoint:
        """
        Returns backend-specific target point.

        :param target_type: Type of the location that should be modified.
        :param target_node_name: Name of the located node.
        :param port_id: id of the port for the statistics disctribution.
        :return: Backend-specific TargetPoint.
        """

    @staticmethod
    @abstractmethod
    def create_bias_correction_command(node: NNCFNode, bias_value: np.ndarray) -> TransformationCommand:
        """
        Creates backend-specific command to update bias value.

        :param node: The node for which bias should be updated.
        :param bias_value: New value for the bias.
        :return: Backend-specific command to update bias value.
        """

    @staticmethod
    @abstractmethod
    def model_extraction_command(
        input_ids: List[Tuple[str, int]], output_ids: List[Tuple[str, int]]
    ) -> TransformationCommand:
        """
        Returns backend-specific command to extract sub-model based on input & output names.

        :param input_ids: List of the input IDs: pairs of node names and correspondent input port ids.
            Each pair denotes the sub-graph beginning.
        :param output_ids: List of the output IDs: pairs of node names and correspondent output port ids.
            Each pair denotes the sub-graph ending.
        :return: Backend-specific TransformationCommand for the model extraction.
        """

    @staticmethod
    @abstractmethod
    def output_insertion_command(nncf_graph: NNCFGraph, target_point: TargetPoint) -> TransformationCommand:
        """
        Returns backend-specific command that inserts output.

        :param nncf_graph: NNCFGraph instance.
        :param target_point: TargetPoint instance.
        :return: Backend-specific command that inserts output.
        """

    @staticmethod
    @abstractmethod
    def mean_statistic_collector(
        channel_axis: int,
        inplace: bool,
        num_samples: Optional[int] = None,
        window_size: Optional[int] = None,
    ) -> TensorStatisticCollectorBase:
        """
        Returns backend-specific mean statistic collector.

        :param channel_axis: Channel axis for the statistics aggregation.
        :param inplace: Whether to calculate statistic inplace or not.
        :param num_samples: Maximum number of samples to collect.
        :param window_size: The maximum size of the samples queue.
        :return: Backend-specific TensorStatisticCollectorBase for the statistics calculation.
        """

    @staticmethod
    @abstractmethod
    def raw_statistic_collector(num_samples: Optional[int] = None) -> TensorStatisticCollectorBase:
        """
        Returns backend-specific raw statistic collector.
        This statistic collector is used for raw data calculation, without aggregating.

        :param num_samples: Maximum number of samples to collect.
        :return: Backend-specific TensorStatisticCollectorBase for the statistics calculation.
        """

    @staticmethod
    @abstractmethod
    def process_model_output(raw_data: OutputType, output_name: str) -> NNCFTensor:
        """
        Returns backend-specific processed output from the model.

        :param raw_data: Backend-specific output from the model.
        :param output_name: Name of the output layer or tensor name.
        :return: Processed output as NNCFTensor.
        """

    @staticmethod
    @abstractmethod
    def get_activation_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> int:
        """
        Returns input port id corresponding to activation input edge for
        the node.
        Supports only nodes that could have bias value.

        :param node: Node of NNCFGraph with bias value.
        :param nncf_graph: NNCFGraph instance with the node.
        :return: target input port id.
        """

    @staticmethod
    @abstractmethod
    def get_bias_value(node: NNCFNode, model: TModel, nncf_graph: NNCFGraph) -> np.ndarray:
        """
        Returns bias value in the NumPy format of provided node.

        :param node: Node of NNCFGraph with bias value.
        :param model: Backend-specific model for the initializer finding.
        :param nncf_graph: NNCFGraph instance with the node.
        :return: Bias value in the NumPy format.
        """

    @staticmethod
    @abstractmethod
    def get_input_name(model: TModel, node_name: str, input_port_id: int) -> str:
        """
        Returns input tensor name for the specific node.

        :param model: Backend-specific model for the initializer finding.
        :param node_name: Name of the backend-specific node.
        :param input_port_id: Port Id for input.
        :return: Input tensor name.
        """

    @staticmethod
    @abstractmethod
    def get_output_name(model: TModel, node_name: str, output_port_id: int) -> str:
        """
        Returns output tensor name for the specific node.

        :param model: Backend-specific model.
        :param node_name: Name of the backend-specific node.
        :param output_port_id: Port Id for output.
        :return: Output tensor name.
        """

    @staticmethod
    @abstractmethod
    def is_quantized_weights(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        """
        Checks whether the node is quantized or not.

        :param node: NNCFNode to check.
        :param nncf_graph: NNCFGraph instance with the node.
        :return: Boolean indicating whether the node has a quantized weights or not.
        """

    @staticmethod
    @abstractmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        """
        Checks whether the node has a bias or not.

        :param node: NNCFNode with the attributes.
        :param nncf_graph: NNCFGraph instance with the node.
        :return: Boolean indicating whether the node has a bias or not.
        """

    @staticmethod
    @abstractmethod
    def remove_fq_from_inputs(model: TModel, nncf_graph: NNCFGraph) -> TModel:
        """
        This method removes the activation Fake Quantize nodes (or Quantize-Dequantize pairs) from the model.
        It's needed for the further bias shift calculation that relates on quantized weights.

        :param model: TModel instance.
        :param nncf_graph: NNCFGraph instance.
        :return: TModel without activation Fake Quantize nodes (or Quantize-Dequantize pairs).
        """



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import List, Tuple, TypeVar

import numpy as np
import torch

import nncf
from nncf.torch.utils import sum_like

GeneralizedTensor = TypeVar("GeneralizedTensor", torch.Tensor, np.ndarray)


class ReferenceBackendType(Enum):
    NUMPY = "numpy"
    TORCH = "torch"


class ReferenceQuantize:
    def __init__(self, backend_type: ReferenceBackendType):
        if backend_type is ReferenceBackendType.NUMPY:
            self.backend = np
        elif backend_type is ReferenceBackendType.TORCH:
            self.backend = torch
        else:
            raise nncf.UnsupportedBackendError("Unknown backend for ReferenceQuantize")

    def _astype(self, tensor: GeneralizedTensor, dtype) -> GeneralizedTensor:
        if self.backend is np:
            return tensor.astype(dtype)
        return tensor.type(dtype)

    def forward(
        self, input_: GeneralizedTensor, input_low: GeneralizedTensor, input_range: GeneralizedTensor, levels: int
    ) -> GeneralizedTensor:
        scale = (levels - 1) / input_range
        output = input_.clip(min=input_low, max=input_low + input_range)
        zero_point = (-input_low * scale).round()
        output -= input_low
        output *= scale
        output -= zero_point
        output = output.round()
        output = output / scale
        return output

    def backward(
        self,
        grad_output: GeneralizedTensor,
        input_: GeneralizedTensor,
        input_low: GeneralizedTensor,
        input_range: GeneralizedTensor,
        output: GeneralizedTensor,
        level_low: int,
        level_high: int,
        is_asymmetric: bool = False,
    ) -> List[GeneralizedTensor]:
        # is_asymmetric is unused, present only to correspond to the CPU signature of calling "backward"
        mask_hi = input_ > (input_low + input_range)
        mask_hi = self._astype(mask_hi, input_.dtype)
        mask_lo = input_ < input_low
        mask_lo = self._astype(mask_lo, input_.dtype)

        mask_in = 1 - mask_hi - mask_lo
        range_sign = np.sign(input_range)
        err = (output - input_) * np.reciprocal(input_range * range_sign)
        grad_range = grad_output * (err * mask_in + range_sign * (level_low / level_high) * mask_lo + mask_hi)
        grad_range = sum_like(grad_range, input_range)

        grad_input = grad_output * mask_in

        grad_low = grad_output * (mask_hi + mask_lo)
        grad_low = sum_like(grad_low, input_low)
        return [grad_input, grad_low, grad_range]

    def tune_range(
        self, input_low: GeneralizedTensor, input_range: GeneralizedTensor, levels: int
    ) -> Tuple[GeneralizedTensor, GeneralizedTensor]:
        input_high = input_range + input_low
        input_low[input_low > 0] = 0
        input_high[input_high < 0] = 0
        n = levels - 1
        scale = n / (input_high - input_low)
        scale = self._astype(scale, input_high.dtype)
        zp = self.backend.round(-input_low * scale)

        new_input_low = self.backend.where(zp < n, zp / (zp - n) * input_high, input_low)
        new_input_high = self.backend.where(zp > 0.0, (zp - n) / zp * input_low, input_high)

        range_1 = input_high - new_input_low
        range_2 = new_input_high - input_low

        mask = self._astype((range_1 > range_2), input_high.dtype)
        inv_mask = abs(1 - mask)

        new_input_low = mask * new_input_low + inv_mask * input_low
        new_input_range = inv_mask * new_input_high + mask * input_high - new_input_low

        return new_input_low, new_input_range


class ReferenceQuantizedFunctions:
    _executor = ReferenceQuantize(backend_type=ReferenceBackendType.TORCH)
    Quantize_forward = _executor.forward
    Quantize_backward = _executor.backward



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Any, Callable, Iterable, List, Optional, Tuple, TypeVar, Union

import nncf
from nncf.common.logging import nncf_logger
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.common.utils.timer import timer
from nncf.data.dataset import Dataset
from nncf.quantization.algorithms.accuracy_control.backend import PreparedModel

TModel = TypeVar("TModel")
TTensor = TypeVar("TTensor")


class IterationCounter:
    """
    A wrapper for counting the passed iterations of iterable objects.
    """

    def __init__(self, iterable):
        self._iterable = iterable
        self._num_iterations = 0

    @property
    def num_iterations(self) -> int:
        return self._num_iterations

    def __iter__(self):
        self._num_iterations = 0
        for x in self._iterable:
            self._num_iterations += 1
            yield x


@dataclass
class MetricResults:
    """
    Results of metrics collection.

    :param metric_value: Aggregated metric value.
    :param values_for_each_item: Metric values for each data item.
    :param preparation_time: Time that it takes to prepare model for validation.
    :param validation_time: Time that it takes to validate model.
    """

    metric_value: float
    values_for_each_item: Union[List[float], List[List[TTensor]]]
    preparation_time: float
    validation_time: float


class Evaluator:
    """
    Evaluator encapsulates a logic to validate model and collect values for each item.
    The value is either calculated metric or model output. This is determined by the
    `Evaluator.is_metric_mode()` method.
    """

    def __init__(
        self, validation_fn: Callable[[Any, Iterable[Any]], Tuple[float, Union[None, List[float], List[List[TTensor]]]]]
    ):
        """
        :param validation_fn: Validation function to validate model.
        """
        self._validation_fn = validation_fn
        self._metric_mode = None
        self._num_passed_iterations = 0
        self._enable_iteration_count = False

    @property
    def num_passed_iterations(self) -> int:
        """
        Number of passed iterations during last validation process if the iteration count is enabled.

        :return: Number of passed iterations during last validation process.
        """

        return self._num_passed_iterations

    def enable_iteration_count(self) -> None:
        """
        Enable the iteration count.
        """
        self._enable_iteration_count = True

    def disable_iteration_count(self) -> None:
        """
        Disable the iteration count.
        """
        self._enable_iteration_count = False

    def is_metric_mode(self) -> Optional[bool]:
        """
        Returns mode of `Evaluator`.

        :return: A boolean indicator where `True` means that the `Evaluator` collects
            metric value for each item and `False` means that the `Evaluator` collects
            logits for each item.
        """
        return self._metric_mode

    def prepare_model(self, model: TModel) -> PreparedModel:
        """
        Prepares model for inference.

        :param model: A model that should be prepared.
        :return: Prepared model for inference.
        """
        backend = get_backend(model)

        if backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.accuracy_control.openvino_backend import OVPreparedModel

            return OVPreparedModel(model)

        if backend == BackendType.ONNX:
            from nncf.quantization.algorithms.accuracy_control.onnx_backend import ONNXPreparedModel

            return ONNXPreparedModel(model)

        raise NotImplementedError(
            f"The `prepare_model_for_inference()` method is not implemented for the {backend} backend."
        )

    def validate_prepared_model(
        self, prepared_model: PreparedModel, dataset: Dataset, indices: Optional[List[int]] = None
    ):
        """
        Validates prepared model for inference.

        :param prepared_model: Prepared model to validate.
        :param dataset: Dataset to validate the model.
        :param indices: Zero-based indices of data items that should be selected from
            the dataset.
        :return: A tuple (metric_value, values_for_each_item) where
            - metric_values: This is a metric for the model.
            - values_for_each_item: If the `Evaluator.is_metric_mode()` condition is true,
                then `values_for_each_item` represents the list of metric value for each item.
                Otherwise, if the condition is false, it represents list of logits for each
                item.
        """
        if self._metric_mode is None:
            self._metric_mode = Evaluator.determine_mode(prepared_model, dataset, self._validation_fn)

        if not self.is_metric_mode() and indices is not None:
            raise ValueError("The `indices` parameter can be used only if Evaluator.is_metric_mode() = True")

        validation_dataset = dataset.get_data(indices)
        if self._enable_iteration_count:
            validation_dataset = IterationCounter(validation_dataset)

        metric, values_for_each_item = self._validation_fn(prepared_model.model_for_inference, validation_dataset)

        self._num_passed_iterations = validation_dataset.num_iterations if self._enable_iteration_count else 0

        if self.is_metric_mode() and values_for_each_item is not None:
            # This casting is necessary to cover the following cases:
            # - np.array(1.0, dtype=np.float32)
            # - np.array([1.0], dtype=np.float32)
            # - torch.tensor(1.0, dtype=torch.float32)
            # - torch.tensor([1.0], dtype=torch.float32)
            # - tf.constant(1.0, dtype=tf.float32
            # - tf.constant([1.0], dtype=tf.float32)
            values_for_each_item = [float(x) for x in values_for_each_item]

        return float(metric), values_for_each_item

    def validate(
        self, model: TModel, dataset: Dataset, indices: Optional[List[int]] = None
    ) -> Tuple[float, Union[None, List[float], List[List[TTensor]]]]:
        """
        Validates model.

        :param model: Model to validate.
        :param dataset: Dataset to validate the model.
        :param indices: Zero-based indices of data items that should be selected from
            the dataset.
        :return: A tuple (metric_value, values_for_each_item) where
            - metric_values: This is a metric for the model.
            - values_for_each_item: If the `Evaluator.is_metric_mode()` condition is true,
                then `values_for_each_item` represents the list of metric value for each item.
                Otherwise, if the condition is false, it represents list of logits for each
                item.
        """
        prepared_model = self.prepare_model(model)
        return self.validate_prepared_model(prepared_model, dataset, indices)

    @staticmethod
    def determine_mode(
        prepared_model: PreparedModel,
        dataset: Dataset,
        validation_fn: Callable[[Any, Iterable[Any]], Tuple[float, Union[None, List[float], List[List[TTensor]]]]],
    ) -> bool:
        """
        Determines mode based on the type of returned value from the
        validation function.

        :param prepared_model: Model to validate.
        :param dataset: Dataset to validate the model.
        :param validation_fn: Validation function to validate model.
        :return: A boolean indicator where `True` means that the `Evaluator` collects
            metric value for each item and `False` means that the `Evaluator` collects
            logits for each item.
        """
        metric_mode = None

        data_item = dataset.get_data([0])

        try:
            metric_value, values_for_each_item = validation_fn(prepared_model.model_for_inference, data_item)
        except Exception:
            metric_mode = False

        if metric_mode is not None:
            return metric_mode

        try:
            metric_value = metric_value if metric_value is None else float(metric_value)
        except Exception as ex:
            raise nncf.InternalError(
                f"Metric value of {type(metric_value)} type was returned from the `validation_fn` "
                "but the float value is expected."
            ) from ex

        convert_to_float_possible = True
        if values_for_each_item is not None:
            try:
                _ = float(values_for_each_item[0])
            except Exception:
                convert_to_float_possible = False

        # Analyze `metric_value` and `values_for_each_item` values:
        # +--------------+----------------------+-------------+
        # | metric_value | values_for_each_item | metric_mode |
        # +--------------+----------------------+-------------+
        # | float        | None                 | True        |
        # +--------------+----------------------+-------------+
        # | float        | List[float]          | True        |
        # +--------------+----------------------+-------------+
        # | float        | List[List[TTensor]]  | False       |
        # +--------------+----------------------+-------------+
        # | None         | None                 | False       |
        # +--------------+----------------------+-------------+
        # | None         | List[float]          | UNEXPECTED  |
        # +--------------+----------------------+-------------+
        # | None         | List[List[TTensor]]  | False       |
        # +--------------+----------------------+-------------+

        metric_mode = False
        if isinstance(metric_value, float) and (values_for_each_item is None or convert_to_float_possible):
            metric_mode = True
        elif values_for_each_item is not None and not isinstance(values_for_each_item[0], list):
            raise nncf.InternalError("Unexpected return value from provided validation function.")

        return metric_mode

    def collect_values_for_each_item_using_prepared_model(
        self, prepared_model: PreparedModel, dataset: Dataset, indices: Optional[List[int]] = None
    ) -> Union[List[float], List[List[TTensor]]]:
        """
        Collects value for each item from the dataset using prepared model for inference.
        If `is_metric_mode()` returns `True` then i-th value is a metric for i-th data item.
        It is an output of the model for i-th data item otherwise.

        :param prepared_model: Model to infer.
        :param dataset: Dataset to collect values.
        :param indices: The zero-based indices of data items that should be selected from
            the dataset.
        :return: Collected values.
        """
        if self._metric_mode:
            # Collect metrics for each item
            values_for_each_item = [
                self._validation_fn(prepared_model.model_for_inference, [data_item])[0]
                for data_item in dataset.get_data(indices)
            ]
        else:
            # Collect outputs for each item
            values_for_each_item = []
            for data_item in dataset.get_inference_data(indices):
                logits = prepared_model(data_item)
                values_for_each_item.append(list(logits.values()))

        self._num_passed_iterations = len(values_for_each_item) if self._enable_iteration_count else 0

        return values_for_each_item

    def collect_values_for_each_item(
        self, model: TModel, dataset: Dataset, indices: Optional[List[int]] = None
    ) -> Union[List[float], List[List[TTensor]]]:
        """
        Collects value for each item from the dataset. If `is_metric_mode()`
        returns `True` then i-th value is a metric for i-th data item. It
        is an output of the model for i-th data item otherwise.

        :param model: A target model.
        :param dataset: Dataset to collect values.
        :param indices: The zero-based indices of data items that should be selected from
            the dataset.
        :return: Collected values.
        """
        prepared_model = self.prepare_model(model)
        return self.collect_values_for_each_item_using_prepared_model(prepared_model, dataset, indices)

    def collect_metric_results(self, model: TModel, dataset: Dataset, model_name: str = "") -> MetricResults:
        """
        Collects metric results.

        :param model: Input model.
        :param dataset: Dataset used to collect metrics.
        :param model_name: Model name.
        :return: Collected metric results.
        """
        nncf_logger.info(f"Validation of {model_name} model was started")

        with timer() as preparation_time:
            prepared_model = self.prepare_model(model)

        with timer() as validation_time:
            metric, values_for_each_item = self.validate_prepared_model(prepared_model, dataset)

        nncf_logger.info(f"Metric of {model_name} model: {metric}")

        if values_for_each_item is None:
            nncf_logger.info(f"Collecting values for each data item using the {model_name} model")
            with timer():
                values_for_each_item = self.collect_values_for_each_item_using_prepared_model(prepared_model, dataset)

        return MetricResults(metric, values_for_each_item, preparation_time(), validation_time())



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from abc import abstractmethod

from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_controller import ElasticityController
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.multi_elasticity_handler import MultiElasticityHandler
from nncf.experimental.torch.nas.bootstrapNAS.training.stage_descriptor import StageDescriptor
from nncf.torch.compression_method_api import PTCompressionAlgorithmController


class BNASTrainingAlgorithm(ABC):
    """
    Base training algorithm for supernet-based NAS methods.
    """

    @abstractmethod
    def step(self) -> None:
        """
        Should be called at the beginning of each training step for activation some Subnet(s).
        """

    @abstractmethod
    def set_stage(self, stage_desc: StageDescriptor) -> None:
        """
        Set a new training stage with parameters from a given stage descriptor

        :param stage_desc: describes parameters of the training stage that should be enabled
        """

    @abstractmethod
    def prepare_for_validation(self) -> None:
        """
        Performs some action on active subnet or supernet before validation. For instance, it can be the batchnorm
        adaptation to achieve the best accuracy on validation.
        """

    @abstractmethod
    def get_total_num_epochs(self) -> int:
        """
        Returns total number of epochs required for the supernet training.

        :return: number of epochs
        """


class BNASTrainingController(PTCompressionAlgorithmController, BNASTrainingAlgorithm, ABC):
    """
    A base class for BootstrapNAS training controllers that provides capabilities for supernet training.
    """

    @property
    @abstractmethod
    def multi_elasticity_handler(self) -> MultiElasticityHandler:
        """
        Gets access to multi elasticity handler to perform some actions with supernet or subnets.

        :return: multi elasticity handler
        """

    @property
    @abstractmethod
    def elasticity_controller(self) -> ElasticityController:
        """
        Gets access to elasticity controller. Usually it's needed for saving its state for further resuming in the
        search part.

        :return: elasticity controller
        """



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import torch

from nncf.common.logging import nncf_logger
from nncf.torch.dynamic_graph.patch_pytorch import register_operator
from nncf.torch.functions import STRound
from nncf.torch.functions import clamp
from nncf.torch.quantization.extensions import QuantizedFunctionsCPU
from nncf.torch.quantization.extensions import QuantizedFunctionsCUDA
from nncf.torch.utils import add_domain


class QuantizeSymmetric(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input_, scale, level_low, level_high, levels):
        input_low = scale * (level_low / level_high)
        input_range = scale - input_low

        if input_.is_cuda:
            if not input_.is_contiguous():
                nncf_logger.debug("input_ is not contiguous!")
                input_ = input_.contiguous()

            # Required to support both torch.amp.autocast and models that perform explicit type casting
            # inside their forward calls.
            if input_.dtype == torch.float16:
                input_low = input_low.type(torch.float16)
                input_range = input_range.type(torch.float16)
            output = QuantizedFunctionsCUDA.get("Quantize_forward")(input_, input_low, input_range, levels)
        else:
            output = QuantizedFunctionsCPU.get("Quantize_forward")(input_, input_low, input_range, levels)

        ctx.save_for_backward(input_, input_low, input_range)
        ctx.levels = levels
        ctx.level_low = level_low
        ctx.level_high = level_high

        return output

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        grad_output = grad_outputs[0]
        input_, input_low, input_range = ctx.saved_tensors
        levels = ctx.levels
        level_low = ctx.level_low
        level_high = ctx.level_high

        if grad_output.is_cuda:
            if not grad_output.is_contiguous():
                nncf_logger.debug("grad_output is not contiguous!")
                grad_output = grad_output.contiguous()

            grad_input, _, grad_scale = QuantizedFunctionsCUDA.get("Quantize_backward")(
                grad_output, input_, input_low, input_range, levels, level_low, level_high
            )
        else:
            grad_input, _, grad_scale = QuantizedFunctionsCPU.get("Quantize_backward")(
                grad_output, input_, input_low, input_range, levels, level_low, level_high, False
            )

        return grad_input, grad_scale, None, None, None


class QuantizeAsymmetric(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input_, input_low, input_range, level_low, level_high, levels):
        if input_.is_cuda:
            if not input_.is_contiguous():
                nncf_logger.debug("input_ is not contiguous!")
                input_ = input_.contiguous()

            # Required to support both torch.amp.autocast and models that perform explicit type casting
            # inside their forward calls.
            if input_.dtype == torch.float16:
                input_low = input_low.type(torch.float16)
                input_range = input_range.type(torch.float16)
            output = QuantizedFunctionsCUDA.get("Quantize_forward")(input_, input_low, input_range, levels)
        else:
            output = QuantizedFunctionsCPU.get("Quantize_forward")(input_, input_low, input_range, levels)

        ctx.save_for_backward(input_, input_low, input_range)
        ctx.levels = levels
        ctx.level_low = level_low
        ctx.level_high = level_high

        return output

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        grad_output = grad_outputs[0]
        input_, input_low, input_range = ctx.saved_tensors
        levels = ctx.levels
        level_low = ctx.level_low
        level_high = ctx.level_high

        if grad_output.is_cuda:
            if not grad_output.is_contiguous():
                nncf_logger.debug("grad_output is not contiguous!")
                grad_output = grad_output.contiguous()

            grad_input, grad_input_low, grad_input_range = QuantizedFunctionsCUDA.get("Quantize_backward")(
                grad_output, input_, input_low, input_range, levels, level_low, level_high
            )
        else:
            grad_input, grad_input_low, grad_input_range = QuantizedFunctionsCPU.get("Quantize_backward")(
                grad_output, input_, input_low, input_range, levels, level_low, level_high, True
            )

        return grad_input, grad_input_low, grad_input_range, None, None, None


def _quantize_autograd_to_range(input_, input_low, input_high, levels):
    input_ = input_ - input_low
    input_range = input_high - input_low
    scale = (levels - 1) / input_range
    output = clamp(input_, low=torch.zeros_like(input_), high=input_range)
    output = output * scale
    output = STRound.apply(output)
    output = output * input_range / (levels - 1) + input_low
    return output


class ExportQuantizeToFakeQuantize(torch.autograd.Function):
    @staticmethod
    def symbolic(
        g, input_, levels, input_low, input_high, output_low, output_high, scale, zero_point, q_min, q_max, ch_axis
    ):
        output = g.op(
            add_domain("FakeQuantize"), input_, input_low, input_high, output_low, output_high, levels_i=levels
        )
        # setType is needed for proper shape inference of custom op on ONNX export. Should work for torch >= 1.14
        output.setType(input_.type())
        return output

    @staticmethod
    def forward(
        ctx, input_, levels, input_low, input_high, output_low, output_high, scale, zero_point, q_min, q_max, ch_axis
    ):
        if ch_axis is not None:
            return torch.fake_quantize_per_channel_affine(input_, scale, zero_point, ch_axis, q_min, q_max)
        return torch.fake_quantize_per_tensor_affine(input_, scale, zero_point, q_min, q_max)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        # backward is not used during export
        return grad_outputs[0]


class ExportQuantizeToONNXQuantDequant(torch.autograd.Function):
    @staticmethod
    def symbolic(g, input_, y_scale, y_zero_point, axis):
        quantized = g.op("QuantizeLinear", input_, y_scale, y_zero_point, axis_i=axis)
        dequantized = g.op("DequantizeLinear", quantized, y_scale, y_zero_point, axis_i=axis)
        return dequantized

    @staticmethod
    def forward(ctx, input_, y_scale, y_zero_point, axis):
        return torch.clone(input_)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        # backward is not used during export
        return grad_outputs[0]


def get_scale_zp_from_input_low_input_high(level_low, level_high, input_low, input_high):
    y_scale = (input_high - input_low) / (level_high - level_low)
    y_zero_point = (level_low * input_high - level_high * input_low) / (input_high - input_low)

    type_ = torch.int8 if level_low < 0 else torch.uint8
    level_low *= torch.ones_like(y_zero_point).to(type_)
    level_high *= torch.ones_like(y_zero_point).to(type_)
    level_low = level_low.to(y_zero_point.device)
    level_high = level_high.to(y_zero_point.device)
    y_zero_point = torch.min(torch.max(level_low, torch.round(y_zero_point).to(type_)), level_high)

    y_scale = torch.squeeze(y_scale)
    y_zero_point = torch.squeeze(y_zero_point)
    return y_scale, y_zero_point


@register_operator()
def symmetric_quantize(input_, levels, level_low, level_high, scale, eps, skip: bool = False):
    if skip:
        return input_
    scale = scale.to(dtype=input_.dtype)
    scale_safe = abs(scale) + eps
    return QuantizeSymmetric.apply(input_, scale_safe, level_low, level_high, levels)


@register_operator()
def asymmetric_quantize(input_, levels, level_low, level_high, input_low, input_range, eps, skip: bool = False):
    if skip:
        return input_
    input_range_safe = abs(input_range) + eps
    input_low_tuned, input_range_tuned = TuneRange.apply(input_low, input_range_safe, levels)
    return QuantizeAsymmetric.apply(input_, input_low_tuned, input_range_tuned, level_low, level_high, levels)


class TuneRange(torch.autograd.Function):
    """
    Makes sure that the zero-point quantum in the quantized domain points exactly to floating point zero,
    e.g. that the input floating point zeroes to the fake quantization operation are translated to output
    floating point zeroes even if we don't use rounding.
    See [docs](../../../docs/compression_algorithms/Quantization.md#asymmetric-quantization) for details.
    """

    @staticmethod
    def forward(ctx, input_low, input_range, levels):
        input_high = input_range + input_low
        input_low_copy = input_low.clone()
        input_low_copy[input_low_copy > 0] = 0
        input_high[input_high < 0] = 0
        n = levels - 1
        # Need a cast here because fp16 division yields fp32 results sometimes
        scale = (n / (input_high - input_low_copy)).to(dtype=input_high.dtype)
        zp = torch.round(-input_low_copy * scale)

        new_input_low = torch.where(zp < n, zp / (zp - n) * input_high, input_low_copy)
        new_input_high = torch.where(zp > 0.0, (zp - n) / zp * input_low_copy, input_high)

        range_1 = input_high - new_input_low
        range_2 = new_input_high - input_low_copy

        mask = (range_1 > range_2).to(input_high.dtype)
        inv_mask = (1 - mask).abs()

        new_input_low = mask * new_input_low + inv_mask * input_low_copy
        new_input_range = inv_mask * new_input_high + mask * input_high - new_input_low

        return new_input_low, new_input_range

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        grad_input_low = grad_outputs[0]
        grad_input_range = grad_outputs[1]
        return grad_input_low, grad_input_range, None


@register_operator()
def decompress(input: torch.Tensor, scale: torch.Tensor, zero_point: torch.Tensor) -> torch.Tensor:
    """
    Decompress the input tensor.

    :param input: An input tensor
    :param scale: A scale tensor
    :param zero_point: A zero point tensor
    :return: The decompressed tensor
    """
    input = input.type(dtype=scale.dtype)
    decompressed_input = (input - zero_point) * scale
    return decompressed_input



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# *WARNING*: Do not run this file directly by `python setup.py install`
# or with any other parameter - this is an outdated and error-prone way
# to install Python packages and causes particular problems with namespace
# packages such as `protobuf`.
# Refer to the table below for well-known and supported alternatives with
# the same behavior:
# +-------------------------------------+---------------------------------------------+
# |           Old command               |                New command                  |
# +-------------------------------------+---------------------------------------------+
# | python setup.py install             | pip install .                               |
# | python setup.py develop             | pip install -e .                            |
# | python setup.py develop --*arg*     | pip install --install-option="*arg*" -e  .  |
# | python setup.py sdist               | python -m build -s                          | <-- using the "build" package
# | python setup.py bdist_wheel         | python -m build -w                          | <-- pypi.org/project/build/
# | python setup.py bdist_wheel --*arg* | python -m build -w -C--global-option=--*arg*|
# +-------------------------------------+---------------------------------------------+
#
# PyPA in general recommends to move away from setup.py and use pyproject.toml
# instead. This doesn't fit us as we currently want to do custom stuff during
# installation such as setting version based on the commit SHA for repo-based
# installs.


import codecs
import glob
import os
import re
import stat
import sys
import sysconfig

import setuptools
from pkg_resources import parse_version
from setuptools import find_packages
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
BKC_SETUPTOOLS_VERSION = "59.5.0"

setuptools_version = parse_version(setuptools.__version__).base_version
if setuptools_version < "43.0.0":
    raise RuntimeError(
        "To properly install NNCF, please install setuptools>=43.0.0, "
        f"while current setuptools version is {setuptools.__version__}. "
        f"Recommended version is {BKC_SETUPTOOLS_VERSION}."
    )

python_version = sys.version_info
if python_version < (3, 8, 0):
    print("Only Python >= 3.8.0 is supported")
    sys.exit(0)

version_string = "{}{}".format(sys.version_info[0], sys.version_info[1])

is_installing_editable = "develop" in sys.argv
is_building_release = not is_installing_editable and "--release" in sys.argv
if "--release" in sys.argv:
    sys.argv.remove("--release")


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if not version_match:
        raise RuntimeError("Unable to find version string.")
    version_value = version_match.group(1)
    if not is_building_release:
        if is_installing_editable:
            return version_value + ".dev0+editable"
        import subprocess  # nosec

        dev_version_id = "unknown_version"
        try:
            repo_root = os.path.dirname(os.path.realpath(__file__))
            dev_version_id = (
                subprocess.check_output(["git", "rev-parse", "--short", "HEAD"], cwd=repo_root)  # nosec
                .strip()
                .decode()
            )
        except subprocess.CalledProcessError:
            pass
        return version_value + f".dev0+{dev_version_id}"

    return version_value


INSTALL_REQUIRES = [
    "jsonschema>=3.2.0",
    "jstyleson>=0.0.2",
    "natsort>=7.1.0",
    "networkx>=2.6, <=3.3",
    "ninja>=1.10.0.post2, <1.12",
    "numpy>=1.19.1, <1.27",
    "openvino-telemetry>=2023.2.0",
    "packaging>=20.0",
    "pandas>=1.1.5,<2.3",
    "psutil",
    "pydot>=1.4.1",
    "pymoo>=0.6.0.1",
    "rich>=13.5.2",
    "scikit-learn>=0.24.0",
    "scipy>=1.3.2",
    "tabulate>=0.9.0",
    "tqdm>=4.54.1",
]

EXTRAS_REQUIRE = {
    "plots": [
        "kaleido>=0.2.1",
        "matplotlib>=3.3.4, <3.6",
        "pillow>=9.0.0",
        "plotly-express>=0.4.1",
    ],
}

with open("{}/README.md".format(here), "r", encoding="utf8") as fh:
    long_description = fh.read()

setup(
    name="nncf",
    version=find_version(os.path.join(here, "nncf/version.py")),
    author="Intel",
    author_email="alexander.kozlov@intel.com",
    description="Neural Networks Compression Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/openvinotoolkit/nncf",
    license="Apache-2.0",
    packages=find_packages(exclude=["tests", "tests.*", "examples", "examples.*", "tools", "tools.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    keywords=[
        "compression",
        "quantization",
        "sparsity",
        "mixed-precision-training",
        "quantization-aware-training",
        "hawq",
        "classification",
        "pruning",
        "object-detection",
        "semantic-segmentation",
        "nas",
        "nlp",
        "bert",
        "transformers",
        "mmdetection",
    ],
    include_package_data=True,
)

path_to_ninja = glob.glob(str(sysconfig.get_paths()["purelib"] + "/ninja*/ninja/data/bin/"))
if path_to_ninja:
    path_to_ninja = str(path_to_ninja[0] + "ninja")
    if not os.access(path_to_ninja, os.X_OK):
        st = os.stat(path_to_ninja)
        os.chmod(path_to_ninja, st.st_mode | stat.S_IEXEC)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import List, Type

import openvino.runtime as ov

from nncf.common.graph import NNCFGraph
from nncf.common.graph.layer_attributes import Dtype
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.openvino.graph.layer_attributes import OVLayerAttributes
from nncf.openvino.graph.metatypes.groups import OPERATIONS_WITH_CONST_PORT_ID
from nncf.openvino.graph.metatypes.openvino_metatypes import OVConvolutionBackpropDataMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVGroupConvolutionBackpropDataMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVGRUSequenceMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVLSTMSequenceMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVMatMulMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import get_node_metatype
from nncf.openvino.graph.metatypes.openvino_metatypes import get_operation_const_op
from nncf.openvino.graph.node_utils import get_weighted_layer_attributes


class GraphConverter:
    """
    Builds the NNCFGraph from an OpenVINO model.
    """

    @staticmethod
    def convert_to_nncf_dtype(ov_type: ov.Type) -> Dtype:
        """
        Converts the primitive types from the OpenVINO domain to the NNCF domain.

        :param ov_dtype: OpenVINO primitive typename.
        :return: NNCF primitive type.
        """
        type_name = ov_type.get_type_name()
        conversion_map = {
            "f16": "float",
            "bf16": "float",
            "f32": "float",
            "f64": "float",
            "i4": "int",
            "i8": "int",
            "i16": "int",
            "i32": "int",
            "i64": "int",
            "u1": "int",
            "u4": "int",
            "u8": "int",
            "u16": "int",
            "u32": "int",
            "u64": "int",
            "boolean": "int",
            "string": "int",
        }
        if type_name not in conversion_map:
            raise NotImplementedError(f"NNCF is not yet supported OpenVINO data type: {type_name}.")
        return Dtype(conversion_map[type_name])

    @staticmethod
    def _filter_weight_input_ports(inputs: List[ov.Input], metatype: Type[OperatorMetatype]) -> List[ov.Input]:
        """
        Specifies the possible weight ports of the OpenVINO node.

        :param inputs: OpenVINO node inputs.
        :param metatype: NNCF meta type which corresponds to operation.
        :return: OpenVINO node inputs that may contain weights.
        """
        if metatype in [OVConvolutionBackpropDataMetatype, OVGroupConvolutionBackpropDataMetatype]:
            return inputs[:2]
        if metatype == OVGRUSequenceMetatype:
            return inputs[:5]
        if metatype == OVLSTMSequenceMetatype:
            return inputs[:6]
        return inputs

    @staticmethod
    def _add_edges_to_nncf_graph(model: ov.Model, graph: NNCFGraph) -> None:
        """
        Adds edges between NNCFNodes to the NNCFGraph.

        :param model: OpenVINO model.
        :param graph: NNCFGraph.
        """
        for op in model.get_ops():
            in_node_id = graph.get_node_by_name(op.get_friendly_name()).node_id
            for output_port_id, out in enumerate(op.outputs()):
                node_vs_target_inputs = defaultdict(list)
                for inp in out.get_target_inputs():
                    node_vs_target_inputs[inp.get_node()].append(inp)

                for out_node, inputs in node_vs_target_inputs.items():
                    tensor_shape = list(out.partial_shape.get_max_shape())
                    output_node_id = graph.get_node_by_name(out_node.get_friendly_name()).node_id
                    nncf_dtype = GraphConverter.convert_to_nncf_dtype(out.get_element_type())

                    parallel_inputs = None
                    if len(inputs) > 1:
                        parallel_inputs = [inp.get_index() for inp in inputs[1:]]

                    graph.add_edge_between_nncf_nodes(
                        from_node_id=in_node_id,
                        to_node_id=output_node_id,
                        tensor_shape=tensor_shape,
                        input_port_id=inputs[0].get_index(),
                        output_port_id=output_port_id,
                        dtype=nncf_dtype,
                        parallel_input_port_ids=parallel_inputs,
                    )

    @staticmethod
    def _add_nncf_node(node: ov.Node, graph: NNCFGraph) -> None:
        """
        Creates NNCFNode from OpenVINO node and adds to the NNCFGraph.

        :param node: OpenVINO node.
        :param graph: NNCFGraph.
        """
        node_type = node.get_type_name()
        metatype = get_node_metatype(node)
        ignored_algorithms = GraphConverter._get_ignored_algorithms(node)
        graph.add_nncf_node(
            node_name=node.get_friendly_name(),
            node_type=node_type,
            node_metatype=metatype,
            ignored_algorithms=ignored_algorithms,
        )

    @staticmethod
    def _get_ignored_algorithms(node: ov.Node) -> List[str]:
        """
        Creates a list of the ignored algorithms corresponding with
        the ignored_algorithms option of add_nncf_node method.

        :param node: OpenVINO node.
        :return: List of the ignored algorithms.
        """
        ignored_algorithms = []
        if "nncf_smooth_quant" in node.get_friendly_name():
            ignored_algorithms.append("ptq_quantization")
        return ignored_algorithms

    @staticmethod
    def create_nncf_graph(model: ov.Model) -> NNCFGraph:
        """
        Creates NNCFGraph from OpenVINO Model.
        All nodes from model which have valid metatype are added to NNCFGraph.
        Then, corresponding edges are added to the NNCFGraph with shape, type, output and input port ids.

        :param model: OpenVINO model.
        :return: NNCFGraph.
        """
        nncf_graph = NNCFGraph()
        visited = set()
        read_value_nodes = [op for op in model.get_ops() if op.get_type_name() == "ReadValue"]
        inference_nodes = model.get_parameters() + read_value_nodes

        while inference_nodes:
            node = inference_nodes[0]
            inference_nodes = inference_nodes[1:]
            if node.get_friendly_name() not in visited:
                GraphConverter._add_nncf_node(node, nncf_graph)
                visited.add(node.get_friendly_name())
                for out in node.outputs():
                    for inp in sorted(out.get_target_inputs(), key=lambda inp: inp.get_node().get_friendly_name()):
                        inference_nodes.append(inp.get_node())

        for node in model.get_ops():
            metatype = get_node_metatype(node)
            # Add nodes from constant subgraphs
            node_name = node.get_friendly_name()
            if node_name not in visited:
                GraphConverter._add_nncf_node(node, nncf_graph)
            # Set const port id
            elif metatype in OPERATIONS_WITH_CONST_PORT_ID:
                const_attrs, act_attrs = {}, {}
                for inp in GraphConverter._filter_weight_input_ports(node.inputs(), metatype):
                    inp_name = inp.get_source_output().get_node().get_friendly_name()
                    if inp_name in visited:
                        continue

                    const_port_id = inp.get_index()
                    const_node = get_operation_const_op(node, const_port_id)
                    if const_node is None:
                        continue

                    if GraphConverter.convert_to_nncf_dtype(const_node.get_element_type()) == Dtype.INTEGER:
                        continue

                    const_attrs[const_port_id] = {
                        "name": const_node.get_friendly_name(),
                        "shape": tuple(const_node.get_output_shape(0)),
                    }

                    if metatype == OVMatMulMetatype:
                        act_port_id = abs(const_port_id - 1)
                        attribute_names = ["transpose_a", "transpose_b"]
                        node_attributes = node.get_attributes()
                        const_transpose_name = attribute_names[const_port_id]
                        const_attrs[const_port_id]["transpose"] = node_attributes[const_transpose_name]
                        act_attrs["transpose"] = node_attributes[attribute_names[act_port_id]]
                    elif metatype == OVGRUSequenceMetatype:
                        node_attributes = node.get_attributes()
                        act_attrs["linear_before_reset"] = node_attributes["linear_before_reset"]

                    if const_attrs or act_attrs:
                        nncf_node = nncf_graph.get_node_by_name(node_name)
                        layer_attributes = get_weighted_layer_attributes(node, metatype, const_attrs)
                        nncf_node.layer_attributes = OVLayerAttributes(const_attrs, layer_attributes, act_attrs)

        GraphConverter._add_edges_to_nncf_graph(model, nncf_graph)
        return nncf_graph



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.api.statistics import Statistics
from nncf.common.utils.helpers import create_table


def _proportion_str(num: int, total_count: int):
    percentage = 100 * (num / max(total_count, 1))
    return f"{percentage:.2f} % ({num} / {total_count})"


class MemoryConsumptionStatistics(Statistics):
    """
    Contains statistics of the memory consumption.
    """

    def __init__(
        self,
        fp32_weight_size: int = 0,
        quantized_weight_size: int = 0,
        max_fp32_activation_size: int = 0,
        max_compressed_activation_size: int = 0,
        weight_memory_consumption_decrease: float = 0.0,
    ):
        """
        Initializes statistics of the memory consumption.

        :param fp32_weight_size: Memory consumption for full-precision weights (Mbyte).
        :param quantized_weight_size: Memory consumption for quantized weights (Mbyte).
        :param max_fp32_activation_size: Max memory consumption for an activation
            tensor in FP32 model (Mbyte).
        :param max_compressed_activation_size: Max memory consumption for an activation
            tensor in compressed model (Mbyte).
        :param weight_memory_consumption_decrease: Memory consumption decrease for weights.
        """
        self.fp32_weight_size = fp32_weight_size
        self.quantized_weight_size = quantized_weight_size
        self.max_fp32_activation_size = max_fp32_activation_size
        self.max_compressed_activation_size = max_compressed_activation_size
        self.weight_memory_consumption_decrease = weight_memory_consumption_decrease

    def to_str(self) -> str:
        memory_consumption_string = create_table(
            header=["Statistic's name", "Value"],
            rows=[
                ["Memory consumption for full-precision weights (Mbyte)", self.fp32_weight_size],
                ["Memory consumption for quantized weights (Mbyte)", self.quantized_weight_size],
                [
                    "Max memory consumption for an activation tensor in FP32 model (Mbyte)",
                    self.max_fp32_activation_size,
                ],
                [
                    "Max memory consumption for an activation tensor in compressed model (Mbyte)",
                    self.max_compressed_activation_size,
                ],
                ["Memory consumption decrease for weights", self.weight_memory_consumption_decrease],
            ],
        )

        pretty_string = f"Statistics of the memory consumption:\n{memory_consumption_string}"
        return pretty_string


class QuantizationConfigurationStatistics(Statistics):
    """
    Contains statistics of the quantization configuration.
    """

    def __init__(self, quantized_edges_in_cfg: int, total_edges_in_cfg: int):
        """
        Initializes statistics of the quantization configuration.

        :param quantized_edges_in_cfg: Number of quantized edges in quantization configuration.
        :param total_edges_in_cfg: Total number of edges in quantization configuration.
        """
        self.quantized_edges_in_cfg = quantized_edges_in_cfg
        self.total_edges_in_cfg = total_edges_in_cfg

    def to_str(self) -> str:
        header = ["Statistic's name", "Value"]
        rows = [
            [
                "Share edges of the quantized data path",
                _proportion_str(self.quantized_edges_in_cfg, self.total_edges_in_cfg),
            ]
        ]
        qc_string = create_table(header, rows)
        pretty_string = f"Statistics of the quantization configuration:\n{qc_string}"
        return pretty_string



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.config.schemata.basic import NUMBER
from nncf.config.schemata.basic import with_attributes

COMPRESSION_LR_MULTIPLIER_PROPERTY = {
    "compression_lr_multiplier": with_attributes(
        NUMBER,
        description="PyTorch only - Used to increase/decrease gradients "
        "for compression algorithms' parameters. The gradients will be multiplied "
        "by the specified value. If unspecified, the gradients will not be adjusted.",
    )
}
BASIC_COMPRESSION_ALGO_SCHEMA = {"type": "object", "required": ["algorithm"]}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import Optional

from timm.layers import Linear
from timm.layers.norm_act import BatchNormAct2d
from timm.layers.norm_act import GroupNormAct
from timm.layers.norm_act import LayerNormAct
from torch import nn

from nncf.torch.nncf_module_replacement import replace_modules_by_nncf_modules


def _copy_parameters(src_module: nn.Module, trg_module: nn.Module):
    """
    Copies parameters of a source module to a target module.
    :param src_module: The source module to copy parameters from.
    :param trg_module: The target module to copy parameters to.
    """
    for name, param in src_module.named_parameters():
        setattr(trg_module, name, deepcopy(param))


def _convert_linear(module: Linear) -> nn.Linear:
    """
    Convert Linear module to torch.nn.Linear.

    param module: The module to convert.
    :return nn.Linear: Converted module.
    """
    with_bias = module.bias is not None
    new_ln = nn.Linear(
        in_features=module.in_features,
        out_features=module.out_features,
        bias=with_bias,
        device=module.weight.device,
        dtype=module.weight.dtype,
    )
    _copy_parameters(module, new_ln)
    return new_ln


def _convert_batch_norm_act_2d(module: BatchNormAct2d) -> nn.Sequential:
    """
    Converts a BatchNormAct2d module to an nn.Sequential module that contains nn.BatchNorm2d,
    followed by dropout and activation functions.

    :param module: The module to convert.
    :return nn.Sequential: A new nn.Sequential module containing nn.BatchNorm2d, dropout, and activation functions.
    """
    new_bn = nn.BatchNorm2d(
        num_features=module.num_features,
        eps=module.eps,
        momentum=module.momentum,
        affine=module.affine,
        track_running_stats=module.track_running_stats,
        device=module.weight.device,
        dtype=module.weight.dtype,
    )
    _copy_parameters(module, new_bn)
    new_bn.running_mean = deepcopy(module.running_mean)
    new_bn.running_var = deepcopy(module.running_var)

    new_drop = deepcopy(module.drop)
    new_act = deepcopy(module.act)
    return nn.Sequential(new_bn, new_drop, new_act)


def _convert_group_norm_act(module: GroupNormAct) -> nn.Sequential:
    """
    Converts a GroupNormAct module to an nn.Sequential module that contains nn.GroupNorm,
    followed by dropout and activation functions.

    :param module: The module to convert.
    :return nn.Sequential: A new nn.Sequential module containing nn.GroupNorm, dropout, and activation functions.
    """
    new_gn = nn.GroupNorm(
        num_groups=module.num_groups,
        num_channels=module.num_channels,
        eps=module.eps,
        affine=module.eps,
        device=module.weight.device,
        dtype=module.weight.dtype,
    )
    _copy_parameters(module, new_gn)
    new_drop = deepcopy(module.drop)
    new_act = deepcopy(module.act)
    return nn.Sequential(new_gn, new_drop, new_act)


def _convert_layer_norm_act(module: LayerNormAct) -> nn.Sequential:
    """
    Converts a LayerNormAct module to an nn.Sequential module that contains nn.LayerNorm,
    followed by dropout and activation functions.

    :param module: The module to convert.
    :return nn.Sequential: A new nn.Sequential module containing nn.LayerNorm, dropout, and activation functions.
    """
    new_norm = nn.LayerNorm(
        normalized_shape=module.normalized_shape,
        eps=module.eps,
        elementwise_affine=module.elementwise_affine,
        device=module.weight.device,
        dtype=module.weight.dtype,
    )
    _copy_parameters(module, new_norm)
    new_drop = deepcopy(module.drop)
    new_act = deepcopy(module.act)
    return nn.Sequential(new_norm, new_drop, new_act)


CONVERT_FN_MAP = {
    BatchNormAct2d: _convert_batch_norm_act_2d,
    GroupNormAct: _convert_group_norm_act,
    LayerNormAct: _convert_layer_norm_act,
    Linear: _convert_linear,
}


def is_timm_custom_module(module: nn.Module):
    """
    Check that module is timm custom module and can be converted.

    :param module: The module.
    :return: `True` if module is custom module, otherwise `False`
    """
    return type(module) in CONVERT_FN_MAP


def convert_timm_custom_modules(module: nn.Module) -> Optional[nn.Module]:
    """
    Replaces the given module with a PyTorch native module if possible.

    :param module: The module to replace.
    :return: The replaced module if replacement is possible, None otherwise.
    """
    module_type = type(module)
    convert_fn = CONVERT_FN_MAP.get(module_type)
    if convert_fn is None:
        raise TypeError(
            f"The type of module {module_type} should be one of the following: {list(CONVERT_FN_MAP.keys())}"
        )
    return convert_fn(module)


def replace_timm_custom_modules_with_torch_native(model: nn.Module) -> nn.Module:
    """
    Replace custom module that can not be operated by NNCF to torch native modules.

    :param model: The target model.
    :return nn.Module: Transformed model.
    """
    model_copy = deepcopy(model)

    model_copy, _ = replace_modules_by_nncf_modules(
        model=model_copy,
        custom_replacer=convert_timm_custom_modules,
        predicate_fn=is_timm_custom_module,
    )

    return model_copy



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base subpackage for NNCF TensorFlow functionality.
"""

import tensorflow
from packaging import version
from packaging.specifiers import SpecifierSet

import nncf
from nncf import nncf_logger
from nncf.common.logging.logger import warn_bkc_version_mismatch
from nncf.version import BKC_TF_SPEC
from nncf.version import STRICT_TF_SPEC

try:
    tensorflow_version = version.parse(version.parse(tensorflow.__version__).base_version)
except:
    nncf_logger.debug("Could not parse tensorflow version")
    tensorflow_version = version.parse("0.0.0")

if tensorflow_version not in SpecifierSet(STRICT_TF_SPEC):
    raise nncf.UnsupportedVersionError(
        f"NNCF only supports tensorflow{STRICT_TF_SPEC}, while current tensorflow version is {tensorflow_version}"
    )
if tensorflow_version not in SpecifierSet(BKC_TF_SPEC):
    warn_bkc_version_mismatch("torch", BKC_TF_SPEC, tensorflow.__version__)


from nncf.common.accuracy_aware_training.training_loop import (
    AdaptiveCompressionTrainingLoop as AdaptiveCompressionTrainingLoop,
)
from nncf.common.accuracy_aware_training.training_loop import (
    EarlyExitCompressionTrainingLoop as EarlyExitCompressionTrainingLoop,
)
from nncf.tensorflow.helpers import create_compressed_model as create_compressed_model
from nncf.tensorflow.helpers.callback_creation import create_compression_callbacks as create_compression_callbacks
from nncf.tensorflow.initialization import register_default_init_args as register_default_init_args
from nncf.tensorflow.pruning.filter_pruning import algorithm as filter_pruning_algorithm

# Required for correct COMPRESSION_ALGORITHMS registry functioning
from nncf.tensorflow.quantization import algorithm as quantization_algorithm
from nncf.tensorflow.sparsity.magnitude import algorithm as magnitude_sparsity_algorithm
from nncf.tensorflow.sparsity.rb import algorithm as rb_sparsity_algorithm



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from nncf.torch.nncf_network import NNCFNetwork


def strip(model: NNCFNetwork, do_copy: bool = True) -> NNCFNetwork:
    """
    Returns the model object with as much custom NNCF additions as possible removed
    while still preserving the functioning of the model object as a compressed model.

    :param do_copy: If True (default), will return a copy of the currently associated model object. If False,
      will return the currently associated model object "stripped" in-place.
    :return: The stripped model.
    """
    return model.nncf.strip(do_copy)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from nncf.telemetry.extractors import CollectedEvent
from nncf.telemetry.extractors import TelemetryExtractor


class CompressionStartedWithQuantizeApi(TelemetryExtractor):
    def extract(self, _: Any) -> CollectedEvent:
        return CollectedEvent(name="compression_started", data="quantize_api")


class CompressionStartedWithQuantizeWithAccuracyControlApi(TelemetryExtractor):
    def extract(self, _: Any) -> CollectedEvent:
        return CollectedEvent(name="compression_started", data="quantize_with_accuracy_control_api")



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from nncf.common.initialization.dataloader import NNCFDataLoader
from nncf.common.utils.api_marker import api
from nncf.config import NNCFConfig
from nncf.config.structures import BNAdaptationInitArgs
from nncf.config.structures import QuantizationRangeInitArgs


class TFInitializingDataLoader(NNCFDataLoader):
    """
    This class wraps the tf.data.Dataset class.

    This is required for proper initialization of certain compression algorithms.
    """

    def __init__(self, data_loader: tf.data.Dataset, batch_size: int):
        self._data_loader = data_loader
        self._batch_size = batch_size

    @property
    def batch_size(self) -> int:
        return self._batch_size

    def __iter__(self):
        return iter(self._data_loader)


@api(canonical_alias="nncf.tensorflow.register_default_init_args")
def register_default_init_args(
    nncf_config: NNCFConfig, data_loader: tf.data.Dataset, batch_size: int, device: str = None
) -> NNCFConfig:
    """
    Register extra structures in the NNCFConfig. Initialization of some
    compression algorithms requires certain extra structures.

    :param nncf_config: An instance of the NNCFConfig class without extra structures.
    :type nncf_config: nncf.NNCFConfig
    :param data_loader: Dataset used for initialization.
    :param batch_size: Batch size used for initialization.
    :param device: Device to perform initialization. If `device` is `None` then the device
        of the model parameters will be used.
    :return: An instance of the NNCFConfig class with extra structures.
    :rtype: nncf.NNCFConfig
    """
    nncf_config.register_extra_structs(
        [
            QuantizationRangeInitArgs(data_loader=TFInitializingDataLoader(data_loader, batch_size), device=device),
            BNAdaptationInitArgs(data_loader=TFInitializingDataLoader(data_loader, batch_size), device=device),
        ]
    )
    return nncf_config



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, List

from nncf import NNCFConfig
from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithm
from nncf.config.extractors import get_bn_adapt_algo_kwargs
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_builder import ElasticityBuilder
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_dim import ElasticityDim
from nncf.experimental.torch.nas.bootstrapNAS.training.progressive_shrinking_controller import (
    ProgressiveShrinkingController,
)
from nncf.experimental.torch.nas.bootstrapNAS.training.scheduler import NASSchedulerParams
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.algo_selector import ZeroCompressionLoss
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.compression_method_api import PTCompressionLoss
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.knowledge_distillation.knowledge_distillation_loss import KnowledgeDistillationLoss
from nncf.torch.model_creation import create_compression_algorithm_builder
from nncf.torch.nncf_network import NNCFNetwork


class PSBuilderStateNames:
    ELASTICITY_BUILDER_STATE = "elasticity_builder_state"
    PROGRESSIVITY_OF_ELASTICITY = "progressivity_of_elasticity"
    BN_ADAPTATION_PARAMS = "bn_adaptation_params"


@PT_COMPRESSION_ALGORITHMS.register("progressive_shrinking")
class ProgressiveShrinkingBuilder(PTCompressionAlgorithmBuilder):
    """
    Determines which modifications should be made to the original FP32 model in
    order to train a supernet using Progressive Shrinking procedure from OFA (https://arxiv.org/abs/1908.09791).
    Operates on an NNCFNetwork object wrapping a target PyTorch model (torch.nn.Module).
    """

    DEFAULT_PROGRESSIVITY = [ElasticityDim.KERNEL, ElasticityDim.DEPTH, ElasticityDim.WIDTH]
    _state_names = PSBuilderStateNames

    def __init__(self, nncf_config: NNCFConfig, should_init: bool = True):
        super().__init__(nncf_config, should_init)
        self._bn_adapt_params = self._algo_config.get("batchnorm_adaptation", {})
        bn_adapt_algo_kwargs = get_bn_adapt_algo_kwargs(nncf_config, self._bn_adapt_params)
        self._bn_adaptation = BatchnormAdaptationAlgorithm(**bn_adapt_algo_kwargs) if bn_adapt_algo_kwargs else None

        default_progressivity = map(lambda x: x.value, self.DEFAULT_PROGRESSIVITY)
        progressivity_of_elasticity = self._algo_config.get("progressivity_of_elasticity", default_progressivity)
        self._progressivity_of_elasticity = list(map(ElasticityDim, progressivity_of_elasticity))
        self._elasticity_builder = ElasticityBuilder(self.config, self.should_init)

        self._lr_schedule_config = self._algo_config.get("lr_schedule", {})

    @staticmethod
    def check_elasticity_dims_consistency(
        available_elasticity_dims: List[ElasticityDim], progressivity_of_elasticity: List[ElasticityDim]
    ) -> None:
        """
        Verifies that progressivity of elasticity is specified for all available elasticity dimensions.

        :param available_elasticity_dims: list of available elasticity dimension
        :param progressivity_of_elasticity: specifies in which order elasticity should be added
        """
        for dim in available_elasticity_dims:
            if dim not in progressivity_of_elasticity:
                raise ValueError(
                    f"Invalid elasticity dimension {dim} specified as available in `elasticity` section."
                    f" This dimension is not part of the progressivity_of_elasticity="
                    f"{progressivity_of_elasticity} which defines order of adding elasticity dimension"
                    f" by going from one training stage to another."
                )

    def initialize(self, model: NNCFNetwork) -> None:
        """
        Initialize model parameters before training

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        """

    def _get_algo_specific_config_section(self) -> Dict:
        return self.config.get("bootstrapNAS", {}).get("training", {})

    def _build_controller(self, model: NNCFNetwork) -> "ProgressiveShrinkingController":
        elasticity_ctrl = self._elasticity_builder.build_controller(model)
        schedule_params = NASSchedulerParams.from_config(self._algo_config.get("schedule", {}))
        compression_loss_func = self._build_compression_loss_function(model)
        return ProgressiveShrinkingController(
            model,
            elasticity_ctrl,
            self._bn_adaptation,
            self._progressivity_of_elasticity,
            schedule_params,
            self._lr_schedule_config,
            compression_loss_func,
        )

    def _build_compression_loss_function(self, model: NNCFNetwork) -> "PTCompressionLoss":
        """
        Create the compression loss. KnowledgeDistillationLoss is returned when knowledge distillation
        algorithm is added to the config. By default, ZeroCompressionLoss is returned.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        """
        compression_builder = create_compression_algorithm_builder(self._algo_config)
        compressed_model = compression_builder.apply_to(model)
        compression_ctrl = compression_builder.build_controller(compressed_model)
        assert type(compression_ctrl.loss) in [
            ZeroCompressionLoss,
            KnowledgeDistillationLoss,
        ], "Currently only knowledge distillation loss is supported."
        return compression_ctrl.loss

    def _get_transformation_layout(self, target_model: NNCFNetwork) -> PTTransformationLayout:
        available_elasticity_dims = self._elasticity_builder.get_available_elasticity_dims()
        self.check_elasticity_dims_consistency(available_elasticity_dims, self._progressivity_of_elasticity)
        return self._elasticity_builder.get_transformation_layout(target_model)

    def _get_state_without_name(self) -> Dict[str, Any]:
        """
        Implementation of get_state that returns state without builder name.

        :return: Returns a dictionary with Python data structures
            (dict, list, tuple, str, int, float, True, False, None) that represents state of the object.
        """
        return {
            self._state_names.ELASTICITY_BUILDER_STATE: self._elasticity_builder.get_state(),
            self._state_names.PROGRESSIVITY_OF_ELASTICITY: [d.value for d in self._progressivity_of_elasticity],
            self._state_names.BN_ADAPTATION_PARAMS: self._bn_adapt_params,
        }

    def _load_state_without_name(self, state_without_name: Dict[str, Any]):
        """
        Implementation of load state that takes state without builder name.

        :param state_without_name: Output of `_get_state_without_name()` method.
        """
        elasticity_builder_state = state_without_name[self._state_names.ELASTICITY_BUILDER_STATE]
        self._elasticity_builder.load_state(elasticity_builder_state)
        progressivity_of_elasticity = state_without_name[self._state_names.PROGRESSIVITY_OF_ELASTICITY]
        # No conflict resolving with the related config options, parameters are overridden by compression state
        self._progressivity_of_elasticity = [ElasticityDim(dim) for dim in progressivity_of_elasticity]
        self._bn_adapt_params = state_without_name[self._state_names.BN_ADAPTATION_PARAMS]
        bn_adapt_algo_kwargs = get_bn_adapt_algo_kwargs(self.config, self._bn_adapt_params)
        self._bn_adaptation = BatchnormAdaptationAlgorithm(**bn_adapt_algo_kwargs) if bn_adapt_algo_kwargs else None



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from nncf.torch.compression_method_api import PTCompressionLoss


# Actually in responsible to lean density to target value
class SparseLoss(PTCompressionLoss):
    def __init__(self, sparse_layers=None, target=1.0, p=0.05):
        super().__init__()
        self._sparse_layers = sparse_layers
        self.target = target
        self.p = p
        self.disabled = False
        self.current_sparsity: float = 0.0
        self.mean_sparse_prob = 0.0

    def set_layers(self, sparse_layers):
        self._sparse_layers = sparse_layers

    def disable(self):
        if not self.disabled:
            self.disabled = True

            for sparse_layer in self._sparse_layers:
                sparse_layer.frozen = True

    def calculate(self) -> torch.Tensor:
        if self.disabled:
            return 0

        params = 0
        loss = 0
        sparse_prob_sum = 0
        for sparse_layer in self._sparse_layers:
            if not self.disabled and sparse_layer.frozen:
                raise AssertionError(
                    "Invalid state of SparseLoss and SparsifiedWeight: mask is frozen for enabled loss"
                )
            if not sparse_layer.frozen:
                sw_loss = sparse_layer.loss()
                params = params + sw_loss.view(-1).size(0)
                loss = loss + sw_loss.sum()
                sparse_prob_sum += torch.sigmoid(sparse_layer.mask).sum()

        self.mean_sparse_prob = (sparse_prob_sum / params).item()
        self.current_sparsity = 1 - loss / params
        return ((loss / params - self.target) / self.p).pow(2)

    @property
    def target_sparsity_rate(self):
        rate = 1 - self.target
        if rate < 0 or rate > 1:
            raise IndexError("Target is not within range(0,1)")
        return rate

    def set_target_sparsity_loss(self, sparsity_level):
        self.target = 1 - sparsity_level


class SparseLossForPerLayerSparsity(SparseLoss):
    def __init__(self, sparse_layers=None, target=1.0, p=0.05):
        super().__init__(sparse_layers, target, p)
        self.per_layer_target = {}
        for sparse_layer in self._sparse_layers:
            self.per_layer_target[sparse_layer] = self.target

    def calculate(self) -> torch.Tensor:
        if self.disabled:
            return 0

        params = 0
        sparse_prob_sum = 0
        sparse_layers_loss = 0
        for sparse_layer in self._sparse_layers:
            if not self.disabled and not sparse_layer.sparsify:
                raise AssertionError(
                    "Invalid state of SparseLoss and SparsifiedWeight: mask is frozen for enabled loss"
                )
            if sparse_layer.sparsify:
                sw_loss = sparse_layer.loss()
                params_layer = sw_loss.view(-1).size(0)
                params += params_layer
                sparse_layers_loss -= torch.abs(sw_loss.sum() / params_layer - self.per_layer_target[sparse_layer])
                sparse_prob_sum += torch.sigmoid(sparse_layer.mask).sum()

        self.mean_sparse_prob = (sparse_prob_sum / params).item()
        return (sparse_layers_loss / self.p).pow(2)

    def set_target_sparsity_loss(self, target, sparse_layer):
        self.per_layer_target[sparse_layer] = 1 - target



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import os
import sys
from abc import ABC
from abc import abstractmethod
from typing import Callable, Optional
from unittest.mock import MagicMock

from nncf import __version__
from nncf.common.logging import nncf_logger
from nncf.definitions import NNCF_CI_ENV_VAR_NAME
from nncf.definitions import NNCF_DEV_ENV_VAR_NAME

NNCFTelemetryStub = MagicMock


class ITelemetry(ABC):
    # For recommendations on proper usage of categories, actions, labels and values, see:
    # https://support.google.com/analytics/answer/1033068

    @abstractmethod
    def start_session(self, category: str, **kwargs):
        """
        Sends a message about starting of a new session.

        :param kwargs: additional parameters
        :param category: the application code
        :return: None
        """

    @abstractmethod
    def send_event(
        self,
        event_category: str,
        event_action: str,
        event_label: str,
        event_value: Optional[int] = None,
        force_send=False,
        **kwargs,
    ):
        """
        Send single event.

        :param event_category: category of the event
        :param event_action: action of the event
        :param event_label: the label associated with the action
        :param event_value: the integer value corresponding to this label
        :param force_send: forces to send event ignoring the consent value
        :param kwargs: additional parameters
        :return: None
        """

    @abstractmethod
    def end_session(self, category: str, **kwargs):
        """
        Sends a message about ending of the current session.

        :param kwargs: additional parameters
        :param category: the application code
        :return: None
        """


def skip_if_raised(func: Callable[..., None]) -> Callable[..., None]:
    """
    For the calls on the decorated function the execution will continue from the statement after the function
    even if an exception was triggered inside the function. The function to be wrapped must return nothing or None.
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            func(*args, **kwargs)

        except Exception as e:
            nncf_logger.debug(f"Skipped calling {func} - internally triggered exception {e}")

    return wrapped


class NNCFTelemetry(ITelemetry):
    MEASUREMENT_ID = "G-W5E9RNLD4H"

    def __init__(self):
        try:
            self._impl = Telemetry(
                app_name="nncf",
                app_version=__version__,
                tid=self.MEASUREMENT_ID,
                backend="ga4",
                enable_opt_in_dialog=False,
                disable_in_ci=True,
            )

        except Exception as e:
            nncf_logger.debug(f"Failed to instantiate telemetry object: exception {e}")

    @skip_if_raised
    def start_session(self, category: str, **kwargs):
        self._impl.start_session(category, **kwargs)

    @skip_if_raised
    def send_event(
        self,
        event_category: str,
        event_action: str,
        event_label: str,
        event_value: Optional[int] = None,
        force_send=False,
        **kwargs,
    ):
        if event_value is None:
            event_value = 1
        self._impl.send_event(event_category, event_action, event_label, event_value, force_send, **kwargs)

    @skip_if_raised
    def end_session(self, category: str, **kwargs):
        self._impl.end_session(category, **kwargs)


try:
    from openvino_telemetry import Telemetry

    telemetry = NNCFTelemetry()
except ImportError:
    nncf_logger.debug("openvino_telemetry package not found. No telemetry will be sent.")
    telemetry = NNCFTelemetryStub()

# Currently the easiest way to disable telemetry in tests. Will break telemetry if pytest is used in NNCF package code.
_IS_IN_PYTEST_CONTEXT = "pytest" in sys.modules

if os.getenv(NNCF_CI_ENV_VAR_NAME) or os.getenv(NNCF_DEV_ENV_VAR_NAME) or _IS_IN_PYTEST_CONTEXT:
    telemetry = NNCFTelemetryStub()



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib
from copy import deepcopy
from enum import Enum
from typing import List, TypeVar

import nncf

TModel = TypeVar("TModel")


class BackendType(Enum):
    TORCH = "Torch"
    TENSORFLOW = "Tensorflow"
    ONNX = "ONNX"
    OPENVINO = "OpenVINO"


def get_available_backends() -> List[BackendType]:
    """
    Returns a list of available backends.

    :return: A list of available backends.
    """
    frameworks = [
        ("torch", BackendType.TORCH),
        ("tensorflow", BackendType.TENSORFLOW),
        ("onnx", BackendType.ONNX),
        ("openvino.runtime", BackendType.OPENVINO),
    ]

    available_backends = []
    for module_name, backend in frameworks:
        try:
            importlib.import_module(module_name)
            available_backends.append(backend)
        except ImportError:
            pass

    return available_backends


def is_torch_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of torch.nn.Module, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of torch.nn.Module, otherwise False.
    """
    import torch

    return isinstance(model, torch.nn.Module)


def is_tensorflow_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of tensorflow.Module, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of tensorflow.Module, otherwise False.
    """
    import tensorflow

    return isinstance(model, tensorflow.Module)


def is_onnx_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of onnx.ModelProto, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of onnx.ModelProto, otherwise False.
    """
    import onnx

    return isinstance(model, onnx.ModelProto)


def is_openvino_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of openvino.runtime.Model, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of openvino.runtime.Model, otherwise False.
    """
    import openvino.runtime as ov

    return isinstance(model, ov.Model)


def is_openvino_compiled_model(model: TModel) -> bool:
    """
    Returns True if the model is an instance of openvino.runtime.CompiledModel, otherwise False.

    :param model: A target model.
    :return: True if the model is an instance of openvino.runtime.CompiledModel, otherwise False.
    """
    import openvino.runtime as ov

    return isinstance(model, ov.CompiledModel)


def get_backend(model: TModel) -> BackendType:
    """
    Returns the NNCF backend name string inferred from the type of the model object passed into this function.

    :param model: The framework-specific model.
    :return: A BackendType representing the correct NNCF backend to be used when working with the framework.
    """
    available_backends = get_available_backends()

    if BackendType.TORCH in available_backends and is_torch_model(model):
        return BackendType.TORCH

    if BackendType.TENSORFLOW in available_backends and is_tensorflow_model(model):
        return BackendType.TENSORFLOW

    if BackendType.ONNX in available_backends and is_onnx_model(model):
        return BackendType.ONNX

    if BackendType.OPENVINO in available_backends and is_openvino_model(model):
        return BackendType.OPENVINO

    raise nncf.UnsupportedBackendError(
        "Could not infer the backend framework from the model type because "
        "the framework is not available or the model type is unsupported. "
        "The available frameworks found: {}.".format(", ".join([b.value for b in available_backends]))
    )


def copy_model(model: TModel) -> TModel:
    """
    Function to create copy of the backend-specific model.

    :param model: the backend-specific model instance
    :return: Copy of the backend-specific model instance.
    """
    model_backend = get_backend(model)
    if model_backend == BackendType.OPENVINO:
        # TODO(l-bat): Remove after fixing ticket: 100919
        return model.clone()
    if model_backend == BackendType.TENSORFLOW:
        # deepcopy and tensorflow.keras.models.clone_model does not work correctly on 2.8.4 version
        from nncf.tensorflow.graph.model_transformer import TFModelTransformer
        from nncf.tensorflow.graph.transformations.layout import TFTransformationLayout

        model = TFModelTransformer(model).transform(TFTransformationLayout())
        return model
    return deepcopy(model)
